import io
import copy
import tokenize
from typeguard import typechecked

from ..constant import AST_POS, SpecialTokens, SPECIAL_CASE
from .. import datatypes as dtX
from ..datatypes import Set, List, Dict, Union, Literal, Optional, Tuple, Any, TokenInfo, MultiLevelPrefix, SpecialTokenMapping
from ..debugger import debugger
from ..lexAnalysis.analyzeConfig import ANALYZE_RULES, AnalyzeEntry, AnalyzeClasses
from ..utils import removeQuotes

@typechecked
def get_tokens_from_code(code: str, isStripped: bool = False) -> List[TokenInfo]:  
  result: List[TokenInfo] = []  
  gen = tokenize.tokenize(io.BytesIO(code.encode('utf-8')).readline)
  next(gen)  # skip the first token, which is always ENCODING

  try:
    for token in gen:
      # debugger.info(token.string, token.type, token.exact_type, token.line)
      result.append({
        "string": token.string,
        "type": token.type,
        "exact_type": token.exact_type,
      })
  except tokenize.TokenError as e:
    pass

  if isStripped:
    result = stripBehindLastLine(result)
  return result

@typechecked
def stripBehindLastLine(lastline: List[TokenInfo]) -> List[TokenInfo]:
  i = len(lastline) - 1
  while i >= 0 and (lastline[i]["exact_type"] == tokenize.NEWLINE or lastline[i]["exact_type"] == tokenize.ENDMARKER):
    i -= 1

  return copy.deepcopy(lastline[:i+1])

@typechecked
def isLeftBracket(t: TokenInfo) -> bool:
  case1 = (t["string"] == "(")
  case2 = (t["string"] == "[")
  case3 = (t["string"] == "{")
  return case1 or case2 or case3

@typechecked
def isRightBracket(t: TokenInfo) -> bool:
  case1 = (t["string"] == ")")
  case2 = (t["string"] == "]")
  case3 = (t["string"] == "}")
  return case1 or case2 or case3

@typechecked
def isBracketMatch(left: TokenInfo, right: TokenInfo) -> bool:
  case1 = (left["string"] == "(" and right["string"] == ")")
  case2 = (left["string"] == "[" and right["string"] == "]")
  case3 = (left["string"] == "{" and right["string"] == "}")
  return case1 or case2 or case3

@typechecked
def checkParenSyntax(tokenList: List[TokenInfo]) -> bool:
  res = True
  signStack: List[TokenInfo] = []
  for tok in tokenList:
    if isLeftBracket(tok):
      signStack.append(tok)
    elif isRightBracket(tok):
      if len(signStack) == 0: # syntax error
        res = False
        break
      elif isBracketMatch(signStack[-1], tok):
        signStack.pop()
      else: # syntax error
        res = False
        break
  return res

@typechecked
def checkParenSymmetry(tokenList: List[TokenInfo]) -> bool:
  res = True
  signStack: List[TokenInfo] = []
  for tok in tokenList:
    if isLeftBracket(tok):
      signStack.append(tok)
    elif isRightBracket(tok):
      if len(signStack) == 0: # syntax error
        res = False
        break
      elif isBracketMatch(signStack[-1], tok):
        signStack.pop()
      else: # syntax error
        res = False
        break
  return res and (len(signStack) == 0)

@typechecked
def groupTokensByDelimiter(tokenList: List[TokenInfo], start_symbol: str, delimiter: str) -> List[List[TokenInfo]]:
  signStack: List[TokenInfo] = []
  contentStack: List[List[TokenInfo]] = []
  cutIdx = 0
  for i, tok in enumerate(tokenList):
    if tok["string"] == delimiter and len(signStack) == 0:
      contentStack.append(copy.deepcopy(tokenList[cutIdx:i]))
      cutIdx = i + 1
    elif isLeftBracket(tok):
      signStack.append(tok)
    elif isRightBracket(tok):
      # We don't need to check the right bracket match, since it is already checked in checkParenSyntax()
      signStack.pop()
  contentStack.append(copy.deepcopy(tokenList[cutIdx:]))
  return contentStack

@typechecked
def inIndexRange(contentStackLen: int, indexRange: List[Optional[int]]):
  idx = contentStackLen - 1
  lowerBound = indexRange[0] if (indexRange[0] is not None) else 0
  if idx < lowerBound:
    # debugger.info("Index out of range !")
    return False
  if indexRange[1] is not None:
    # debugger.info("Index out of range !!", idx)
    return idx < indexRange[1]
  else:
    # debugger.info("Index out of range !!!")
    return True

@typechecked
def lastLevelPrefixMatch(tokenList: List[TokenInfo], lastLevelPrefix: Optional[List[str]] = None) -> bool:
  if lastLevelPrefix is None:
    return True
  tlLen = len(tokenList)
  llpLen = len(lastLevelPrefix)
  if tlLen != llpLen:
    return False
  for i in range(llpLen):
    if tokenList[i]["string"] != lastLevelPrefix[i]:
      return False
  return True


@typechecked
def multiLevelPrefixMatch(tokenList: List[TokenInfo], multiLevelPrefixes: List[MultiLevelPrefix], lastLevelPrefix: Optional[List[str]] = None) -> bool:
  """
  Input: 
    - A list of tokens like `["(", "ignore_index", "=", "true", ",", "[", "df1", ","]`
    - A list of multi-level prefixes like
    ```[{
      "start_symbol": "(",
      "delimiter": ",",
      "index_range": [1, 2],
      }, {
      "start_symbol": "[",
      "delimiter": ",",
      "index_range": [None, None]
    }]```
    - The prefix of last level like: `[]`

  Output: A boolean value indicating whether the tokens match the multi-level prefix.
  """
  # check the parenthesis syntax
  if not checkParenSyntax(tokenList):
    return False

  res = True
  contentStack: List[List[TokenInfo]] = []
  resToken: List[TokenInfo] = copy.deepcopy(tokenList)
  for mp in multiLevelPrefixes:
    # debugger.info(mp)
    mp_ss = mp["start_symbol"]
    mp_d = mp["delimiter"]
    mp_ir = mp["index_range"]
    # Check if the start symbol matches
    if len(resToken) == 0 or resToken[0]["string"] != mp_ss:
      res = False
      # debugger.info("Start symbol not match")
      break
    # Check symmetry of parenthesis
    if checkParenSymmetry(resToken):
      # Which means we are not in that level
      res = False
      break
    resToken = resToken[1:]
    # Split and group the tokens by delimiter
    contentStack = groupTokensByDelimiter(resToken, mp_ss, mp_d)
    # debugger.info(contentStack, mp_ir, resToken)
    if not inIndexRange(len(contentStack), mp_ir):
      res = False
      # debugger.info("Index range not match")
      break
    resToken = contentStack[-1]
  
  # Check the last level prefix
  if not lastLevelPrefixMatch(resToken, lastLevelPrefix):
    res = False

  return res


@typechecked
def keepSigBody(lastline_tokens: List[TokenInfo], sig_name: str, pandas_alias: Optional[str] = None, allDfInfo = None) -> Tuple[List[TokenInfo], List[TokenInfo]]:
  if len(lastline_tokens) < 4:
    raise ValueError("[keepSigBody] lastline_tokens should have at least 4 tokens")
  case1 = pandas_alias is None
  case2 = allDfInfo is None
  if (case1 and case2) or (not case1 and not case2):
    raise ValueError("[keepSigBody] pandas_alias and allDfInfo are both None or not None")
  
  all_df_names = list(allDfInfo.keys()) if allDfInfo is not None else []
  prefix: List[TokenInfo] = []
  findBody = False
  res: List[TokenInfo] = copy.deepcopy(lastline_tokens)
  for i in range(len(res)-4, -1, -1):
    if res[i]["string"] == pandas_alias and res[i+1]["string"] == "." and res[i+2]["string"] == sig_name:
      findBody = True
      prefix = res[:i+3]
      res = res[i+3:]
      break
    elif res[i]["string"] in all_df_names and res[i+1]["string"] == "." and res[i+2]["string"] == sig_name:
      findBody = True
      prefix = res[:i+3]
      res = res[i+3:]
      break
    elif res[i]["string"] == "]" and res[i+1]["string"] == "." and res[i+2]["string"] == sig_name:
      # e.g. df["A"].isin(
      if i < 3:
        continue
      if res[i-3]["string"] not in all_df_names:
        continue
      col_in_df = allDfInfo.get(res[i-3]["string"], {"columnNameList": []}).get("columnNameList", [])
      if res[i-2]["string"] == "[" and removeQuotes(res[i-1]["string"]) in col_in_df:
        findBody = True
        prefix = res[:i+3]
        res = res[i+3:]
        break
    elif res[i]["string"] == "str" and res[i+1]["string"] == "." and res[i+2]["string"] == sig_name:
      # e.g. df["A"].str.replace(
      if i < 5:
        continue
      if res[i-5]["string"] not in all_df_names:
        continue
      col_in_df = allDfInfo.get(res[i-5]["string"], {"columnNameList": []}).get("columnNameList", [])
      if res[i-4]["string"] == "[" and removeQuotes(res[i-3]["string"]) in col_in_df and res[i-2]["string"] == "]" and res[i-1]["string"] == ".":
        findBody = True
        prefix = res[:i+3]
        res = res[i+3:]
        break
  if not findBody:
    raise ValueError("[keepSigBody] Cannot find the body of signature")
  return res, prefix

@typechecked
def isPandasAlias(key: str, value: str, pandasAlias: str) -> bool:
  return (key == SpecialTokens.PANDAS) and (value == pandasAlias)

@typechecked
def isDfName(key: str, value: str, all_df_names: List[str]) -> bool:
  return (key == SpecialTokens.DF) and (value in all_df_names)

@typechecked
def isLikeColIdx(key: str, value_type: int) -> bool:
  return (key == SpecialTokens.COLIDX) and (value_type == tokenize.STRING)

@typechecked
def isLikeCellValue(key: str, value_type: int) -> bool:
  return (key == SpecialTokens.CELL_VALUE) and (any(value_type == t for t in [tokenize.STRING, tokenize.NUMBER]))

@typechecked
def isOp(key: str, value_type: int, value: str) -> bool:
  return (key == SpecialTokens.OP) and (value_type == tokenize.OP) and value != "="

@typechecked
def specialTokenMatch(prefix_i: int, lastline_i: int, prefixList: List[str], lastline_tokens: List[TokenInfo], pandasAlias: str, all_df_names: List[str]) -> Tuple[bool, int, int]:
  case1 = isPandasAlias(prefixList[prefix_i], lastline_tokens[lastline_i]["string"], pandasAlias)
  case2 = isDfName(prefixList[prefix_i], lastline_tokens[lastline_i]["string"], all_df_names)
  case3 = isLikeColIdx(prefixList[prefix_i], lastline_tokens[lastline_i]["type"])
  # Whether or not the string is a valid column index leaved to following check
  case4 = prefixList[prefix_i] == SpecialTokens.COLIDX_LIST
  # Whether or not the value is valid cell value leaved to following check
  case5 = isLikeCellValue(prefixList[prefix_i], lastline_tokens[lastline_i]["type"])
  case6 = isOp(prefixList[prefix_i], lastline_tokens[lastline_i]["type"], lastline_tokens[lastline_i]["string"])
  
  if case1 or case2 or case3 or case5 or case6:
    return True, prefix_i + 1, lastline_i + 1
  elif case4:
    current_is: Literal["colidx", "comma"] = "colidx"
    while prefix_i < len(prefixList) and lastline_i < len(lastline_tokens):
      if current_is == "colidx" and isLikeColIdx(SpecialTokens.COLIDX, lastline_tokens[lastline_i]["type"]):
        lastline_i += 1
        current_is = "comma"
      elif current_is == "comma" and lastline_tokens[lastline_i]["string"] == ",":
        lastline_i += 1
        current_is = "colidx"
      elif current_is == "comma" and lastline_tokens[lastline_i]["string"] == "[":
        break
      else:
        lastline_i += 1
        case4 = False
        break
    
    return case4, prefix_i + 1, lastline_i
  else:
    return False, prefix_i + 1, lastline_i + 1

@typechecked
def prefixMatch(lastline_tokens: List[TokenInfo], prefixList: List[str], pandasAlias: str, all_df_names: List[str]) -> Tuple[bool, List[SpecialTokenMapping]]:
  specialTokenValues: List[SpecialTokenMapping] = []
  if len(lastline_tokens) < len(prefixList):
    return False, specialTokenValues
  
  lastline_tokens_copy = copy.deepcopy(lastline_tokens)
  prefixList_copy = copy.deepcopy(prefixList)
  lastline_tokens_copy.reverse()
  prefixList_copy.reverse()
  res: bool = True
  
  prefix_i = 0
  lastline_i = 0
  while prefix_i < len(prefixList_copy) and lastline_i < len(lastline_tokens_copy):
    if prefixList_copy[prefix_i] == lastline_tokens_copy[lastline_i]["string"]:
      prefix_i += 1
      lastline_i += 1
      continue
    res, new_prefix_i, new_lastline_i = specialTokenMatch(prefix_i, lastline_i, prefixList_copy, lastline_tokens_copy, pandasAlias, all_df_names)
    if not res:
      break
    else:
      v = lastline_tokens_copy[new_lastline_i-1:lastline_i-1:-1] if lastline_i >= 1 else lastline_tokens_copy[new_lastline_i-1::-1]
      specialTokenValues.append(SpecialTokenMapping(
        key=prefixList_copy[prefix_i],
        value="".join(list(map(lambda x: x["string"], v))),
      ))
      prefix_i = new_prefix_i
      lastline_i = new_lastline_i
  
  specialTokenValues.reverse()
  return res, specialTokenValues

@typechecked
def matchRulesPDFuncSig(lastline_tokens: List[TokenInfo], pandasAlias: str, sigName: str) -> dict:
  final_rule = {
    "need_obj": AST_POS.INVALID,
    "df_info_type": "",
    "full_name": "",
    "is_trivial": False,
  }
  # At least, like: `pd.concat(`
  if len(lastline_tokens) < 4:
    return final_rule
  lastline_tokens_body, _ = keepSigBody(lastline_tokens, sigName, pandas_alias=pandasAlias)

  funcConfig: Optional[dict] = ANALYZE_RULES.get(AnalyzeClasses.IN_PD_FUNC_SIG)
  if funcConfig is None:
    raise ValueError("[matchRulesPDFuncSig] No analyze rules for IN_PD_FUNC_SIG")
  funcConfig = funcConfig.get(sigName)
  if funcConfig is None:
    return final_rule
  funcRules = funcConfig.get(AnalyzeEntry.RULES)
  if funcRules is None:
    return final_rule
  
  rule: Optional[dict] = None
  for _rule in funcRules:
    rule = _rule
    mlPrefix: List[MultiLevelPrefix] = rule.get(AnalyzeEntry.ML_PREFIX)
    llPrefix: Optional[List[str]] = rule.get(AnalyzeEntry.LL_PREFIX)
    if multiLevelPrefixMatch(lastline_tokens_body, mlPrefix, llPrefix):
      final_rule["need_obj"] = rule.get(AnalyzeEntry.NEED_OBJ)
      final_rule["df_info_type"] = rule.get(AnalyzeEntry.DF_INFO_TYPE)
      final_rule["is_trivial"] = rule.get(AnalyzeEntry.IS_TRIVIAL)
      final_rule["full_name"] = funcConfig.get(AnalyzeEntry.FULL_NAME)
      break
  return final_rule

@typechecked
def matchRulesGlobal(lastline_tokens: List[TokenInfo], pandasAlias: str, allDfInfo) -> Tuple[dict, List[SpecialTokenMapping]]:
  final_rule = {
    "need_obj": AST_POS.INVALID,
    "df_info_type": "",
    "full_name": "",
    "is_trivial": False,
  }
  matchDetail: List[SpecialTokenMapping] = []
  # 1. find all configuratons for global rules
  gblCfg: Dict[str, Dict[str, Any]] = ANALYZE_RULES.get(AnalyzeClasses.GLOBAL)
  if gblCfg is None:
    raise ValueError("[matchRulesGlobal] No analyze rules for GLOBAL")
  # 2. traverse all configurations and all rules and find the first match
  cfgIds = list(gblCfg.keys())
  for id in cfgIds:
    found = False
    oneCfg = gblCfg.get(id)
    if (oneCfg is None) or (oneCfg.get(AnalyzeEntry.RULES) is None):
      # debugger.warning(f"[matchRulesGlobal] No analyze rules for {id}")
      continue
    rules: List[Dict[str, Any]] = oneCfg.get(AnalyzeEntry.RULES)
    for r in rules:
      mlPrefix: Optional[List[MultiLevelPrefix]] = r.get(AnalyzeEntry.ML_PREFIX)
      if mlPrefix is not None:
        # debugger.warning(f"[matchRulesGlobal] Multi-level prefix is currently not supported for global rules")
        pass
      llPrefix: List[str] = r.get(AnalyzeEntry.LL_PREFIX)
      isMatch, matchDetail = prefixMatch(lastline_tokens, llPrefix, pandasAlias, list(allDfInfo.keys()))
      if isMatch:
        final_rule["need_obj"] = r.get(AnalyzeEntry.NEED_OBJ)
        final_rule["df_info_type"] = r.get(AnalyzeEntry.DF_INFO_TYPE)
        final_rule["full_name"] = oneCfg.get(AnalyzeEntry.FULL_NAME)
        final_rule["is_trivial"] = r.get(AnalyzeEntry.IS_TRIVIAL)
        found = True
        break
    if found:
      break
  return final_rule, matchDetail

@typechecked
def matchRulesDfFuncSig(lastline_tokens: List[TokenInfo], allDfInfo, sigName: str) -> Tuple[dict, str]:
  final_rule = {
    "need_obj": AST_POS.INVALID,
    "df_info_type": "",
    "full_name": "",
    "is_trivial": False,
  }
  dfName: str = ""
  # At least, like: `df.merge(`
  if len(lastline_tokens) < 4:
    return final_rule, dfName
  lastline_tokens_body, prefix = keepSigBody(lastline_tokens, sigName, allDfInfo=allDfInfo)
  dfName = prefix[-3]["string"]

  funcConfig: Optional[dict] = ANALYZE_RULES.get(AnalyzeClasses.IN_DF_METHOD_SIG)
  if funcConfig is None:
    raise ValueError("[matchRulesDfFuncSig] No analyze rules for IN_DF_METHOD_SIG")
  funcConfig = funcConfig.get(sigName)
  if funcConfig is None:
    return final_rule, dfName
  funcRules = funcConfig.get(AnalyzeEntry.RULES)
  if funcRules is None:
    return final_rule, dfName
  
  rule: Optional[dict] = None
  for _rule in funcRules:
    rule = _rule
    mlPrefix: List[MultiLevelPrefix] = rule.get(AnalyzeEntry.ML_PREFIX)
    llPrefix: Optional[List[str]] = rule.get(AnalyzeEntry.LL_PREFIX)
    if multiLevelPrefixMatch(lastline_tokens_body, mlPrefix, llPrefix):
      final_rule["need_obj"] = rule.get(AnalyzeEntry.NEED_OBJ)
      final_rule["df_info_type"] = rule.get(AnalyzeEntry.DF_INFO_TYPE)
      final_rule["full_name"] = funcConfig.get(AnalyzeEntry.FULL_NAME)
      final_rule["is_trivial"] = rule.get(AnalyzeEntry.IS_TRIVIAL)
      break
  return final_rule, dfName


@typechecked
def matchRulesSeFuncSig(lastline_tokens: List[TokenInfo], pandasAlias: str, allDfInfo, sigName: str) -> Tuple[dict, str, str]:
  final_rule = {
    "need_obj": AST_POS.INVALID,
    "df_info_type": "",
    "full_name": "",
    "is_trivial": False,
  }
  dfName: str = ""
  colName: str = ""
  # Assume `df["A"].isin(`
  if len(lastline_tokens) < 7:
    return final_rule, dfName, colName
  lastline_tokens_body, prefix = keepSigBody(lastline_tokens, sigName, allDfInfo=allDfInfo)
  dfName = prefix[-6]["string"]
  colName = removeQuotes(prefix[-4]["string"])

  funcConfig: Optional[dict] = ANALYZE_RULES.get(AnalyzeClasses.IN_COL_METHOD_SIG)
  if funcConfig is None:
    raise ValueError("[matchRulesSeFuncSig] No analyze rules for IN_COL_METHOD_SIG")
  funcConfig = funcConfig.get(sigName)
  if funcConfig is None:
    return final_rule, dfName, colName
  funcRules = funcConfig.get(AnalyzeEntry.RULES)
  if funcRules is None:
    return final_rule, dfName, colName
  
  rule: Optional[dict] = None
  for _rule in funcRules:
    rule = _rule
    mlPrefix: List[MultiLevelPrefix] = rule.get(AnalyzeEntry.ML_PREFIX)
    llPrefix: Optional[List[str]] = rule.get(AnalyzeEntry.LL_PREFIX)
    if multiLevelPrefixMatch(lastline_tokens_body, mlPrefix, llPrefix):
      final_rule["need_obj"] = rule.get(AnalyzeEntry.NEED_OBJ)
      final_rule["df_info_type"] = rule.get(AnalyzeEntry.DF_INFO_TYPE)
      final_rule["full_name"] = funcConfig.get(AnalyzeEntry.FULL_NAME)
      final_rule["is_trivial"] = rule.get(AnalyzeEntry.IS_TRIVIAL)
      break
  return final_rule, dfName, colName

def matchRulesSeStrFuncSig(lastline_tokens: List[TokenInfo], pandasAlias: str, allDfInfo, sigName: str) -> Tuple[dict, str, str]:
  final_rule = {
    "need_obj": AST_POS.INVALID,
    "df_info_type": "",
    "full_name": "",
    "is_trivial": False,
  }
  dfName: str = ""
  colName: str = ""
  # At least, like: `df["A"].str.replace(`
  if len(lastline_tokens) < 9:
    return final_rule, dfName, colName
  lastline_tokens_body, prefix = keepSigBody(lastline_tokens, sigName, allDfInfo=allDfInfo)
  dfName = prefix[-8]["string"]
  colName = removeQuotes(prefix[-6]["string"])

  funcConfig: dict = ANALYZE_RULES.get(AnalyzeClasses.IN_COL_STR_METHOD_SIG)
  funcConfig = funcConfig.get(sigName, {})
  funcRules: list = funcConfig.get(AnalyzeEntry.RULES, [])
  
  rule: dict = {}
  for _rule in funcRules:
    rule = _rule
    mlPrefix: List[MultiLevelPrefix] = rule.get(AnalyzeEntry.ML_PREFIX)
    llPrefix: Optional[List[str]] = rule.get(AnalyzeEntry.LL_PREFIX)
    if multiLevelPrefixMatch(lastline_tokens_body, mlPrefix, llPrefix):
      final_rule["need_obj"] = rule.get(AnalyzeEntry.NEED_OBJ)
      final_rule["df_info_type"] = rule.get(AnalyzeEntry.DF_INFO_TYPE)
      final_rule["full_name"] = funcConfig.get(AnalyzeEntry.FULL_NAME)
      final_rule["is_trivial"] = rule.get(AnalyzeEntry.IS_TRIVIAL)
      break
  return final_rule, dfName, colName


@typechecked
def getPandasAlias(code: List[List[str]]) -> str:
  default_alias = "pandas"
  target = "import pandas as "
  res = ""
  for block in code:
    for line in block:
      idx = line.find(target)
      if idx >= 0:
        res = line[len(target):].strip()
  if len(res) > 0:
    return res
  else:
    return default_alias

@typechecked
def isComment(lastline_tokens: List[TokenInfo]):
  yes = False
  for t in lastline_tokens:
    if t["exact_type"] == tokenize.COMMENT:
      yes = True
      break
  return yes

@typechecked
def getSTFromMatchDetail(matchDetail: List[SpecialTokenMapping], stKey: str, how: Literal["any", "all"] = "any") -> Union[None, str, List[str]]:
  # get special token value from matchDetail
  res: List[str] = []
  for m in matchDetail:
    if m["key"] == stKey:
      res.append(m["value"])
    if how == "any" and len(res) > 0:
      break
  return None if len(res) == 0 else res if how == "all" else res[0]

@typechecked
def isDfNamePrefix(tok: str, allDfInfo) -> bool:
  # example: tok = df, actual name = "df_name"
  dfNames = list(allDfInfo.keys())
  for dn in dfNames:
    if dn.startswith(tok):
      return True
  return False

@typechecked
def isColNamePrefix(tok: str, tableLvInfo: dtX.TableLevelInfoT, dfName: str) -> bool:
  # example: tok = col, actual name = "col_name"
  colNames = tableLvInfo.get(dfName, {"columnNameList": []}).get("columnNameList", [])
  for cn in colNames:
    if cn.startswith(tok):
      return True
  return False

@typechecked
def specialCaseDetect(lastline_tokens: List[TokenInfo], pandasAlias: str, tableLvInfo: dtX.TableLevelInfoT) -> int:
  case3 = len(lastline_tokens) >= 1 and isDfNamePrefix(lastline_tokens[-1]["string"], tableLvInfo)

  case4 = len(lastline_tokens) >= 2 and lastline_tokens[-1]["string"] == "[" and lastline_tokens[-2]["exact_type"] == tokenize.NAME and lastline_tokens[-2]["string"] in tableLvInfo

  case5 = len(lastline_tokens) >= 3 and lastline_tokens[-1]["string"] in ["'", '"'] and lastline_tokens[-2]["string"] == "[" and lastline_tokens[-3]["exact_type"] == tokenize.NAME and lastline_tokens[-3]["string"] in tableLvInfo

  case6 = len(lastline_tokens) >= 4 and lastline_tokens[-4]["exact_type"] == tokenize.NAME and lastline_tokens[-4]["string"] in tableLvInfo and lastline_tokens[-3]["string"] == "[" and lastline_tokens[-2]["string"] in ["'", '"'] and lastline_tokens[-1]["exact_type"] == tokenize.NAME and isColNamePrefix(lastline_tokens[-1]["string"], tableLvInfo, lastline_tokens[-4]["string"])

  case7 = len(lastline_tokens) >= 2 and lastline_tokens[-2]["string"] in ["'", '"'] and lastline_tokens[-1]["exact_type"] == tokenize.NAME

  # debugger.info(f"[specialCaseDetect] {[(t['string'], t['exact_type']) for t in lastline_tokens]}, {case3}, {case4}, {case5}, {case6}, {case7}")
  if case6:
    return SPECIAL_CASE.DF_SELECT_COL_2
  elif case5:
    return SPECIAL_CASE.DF_SELECT_COL_1
  elif case4:
    return SPECIAL_CASE.DF_SELECT
  elif case3:
    return SPECIAL_CASE.DF_NAME_PREFIX
  elif case7:
    return SPECIAL_CASE.DF_SELECT_COL_3
  
  return SPECIAL_CASE.NONE


import pandas as pd

DEFAULT_ROW_SAMPLE_SIZE = 15

def selectRowsWithinRange(df: pd.DataFrame, column: str, rangeSelect: list):
  lower_op = None
  upper_op = None
  evalStr = ""
  for i, rs in enumerate(rangeSelect):
    if rs["startOpen"]:
      lower_op = '>'
    else:
      lower_op = '>='
    if rs["endOpen"]:
      upper_op = '<'
    else:
      upper_op = '<='
    if i > 0:
      evalStr += " | "
    evalStr += f'(`{column}` {lower_op} {rs["start"]} & `{column}` {upper_op} {rs["end"]})'
  mask = df.eval(evalStr)
  return df[mask]

def convertBinned(numVC: pd.Series, true_min):
  """
  numVC has interval index from binned value counts. Replace far left low 
  with true min because pandas cuts below min
  """
  d = pd.DataFrame({
    "low": numVC.index.left,
    "high": numVC.index.right,
    "count": numVC.values
  })
  d = d.reset_index().rename(columns={"index": "bucket"})
  d_dict = d.to_dict('records')

  if len(d_dict) > 0:
    d_dict[0]['low'] = true_min
  
  return d_dict

def getQuantBinnedData(colData: pd.Series, n: int = 20):
  vc = colData.value_counts(bins=min(n, colData.nunique()), sort=False)
  return vc

def getTempBinnedData(colData: pd.Series, n: int = 20):
    vc = (colData.astype("int64")//1e9).value_counts(bins=min(n, colData.nunique()), sort=False)
    return vc

"""
export function isTimeColumn(dtype: string) {
  return ["datetime64", "datetime64[ns]"].includes(dtype);
}

export function isNumColumn(dtype: string) {
  const INT_TYPE = ['int', 'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64', 'Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32', 'UInt64'];
  const FLOAT_TYPE = ['float', 'float_', 'float16', 'float32', 'float64'];
  const NUM_TYPE = [...INT_TYPE, ...FLOAT_TYPE];
  return NUM_TYPE.includes(dtype) || isTimeColumn(dtype);
}
"""

def isTimeColumn(dtype: str):
  return dtype in ['datetime64', 'datetime64[ns]']

def isNumColumn(dtype: str):
  INT_TYPE = ['int', 'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64', 'Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32', 'UInt64']
  FLOAT_TYPE = ['float', 'float_', 'float16', 'float32', 'float64']
  NUM_TYPE = INT_TYPE + FLOAT_TYPE
  return dtype in NUM_TYPE or isTimeColumn(dtype)


import re
from typing import Optional, List, Dict, Tuple, Any
from typeguard import typechecked


@typechecked
def keepSuffix(srcStr: str, newStr: str) -> str:
  # Example: keepSuffix("abcdefg", "efghijk") => "hijk"
  matchSubLen = -1
  for i in range(min(len(srcStr), len(newStr))):
    sublen = i + 1
    srcStrSub = srcStr[(len(srcStr)-sublen):len(srcStr)]
    newStrSub = newStr[0:sublen]
    if srcStrSub == newStrSub:
      matchSubLen = sublen
  
  if matchSubLen == -1:
    return newStr
  else:
    return newStr[matchSubLen:]

@typechecked
def removeBackTick(text: str) -> str:
  # Example: removeBackTick("`hello`") => "hello"
  if text.startswith("```") and text.endswith("```"):
    return text[3:-3]
  elif text.startswith("`") and text.endswith("`"):
    return text[1:-1]
  else:
    return text
  
@typechecked
def parseReCodeObj(codeObj: str, lastLineCode: str) -> List[str]:
  pres: List[str] = []
  lines = codeObj.split("\n")
  for line in lines:
    # Use regular expression to extract things like "1. " or "2. "
    searchres = re.search(r"\d+\.[ ]*", line)
    if searchres:
      cutidx = searchres.end()
      realCodeObj = removeBackTick(line[cutidx:])
      pres.append(keepSuffix(lastLineCode, realCodeObj))
  return pres

@typechecked
def parseReCodeObjWithExplanation(codeObj: str, lastLineCode: str) -> Tuple[List[str], List[str]]:
  pres: List[str] = []
  pexp: List[str] = []
  lines = codeObj.split("\n")
  for line in lines:
    # Use regular expression to extract things like "1. " or "2. "
    searchres = re.search(r"\d+\.[ ]*", line)
    if searchres:
      cutidx = searchres.end()
      codeAndExp = line[cutidx:].split("|")
      code = codeAndExp[0].rstrip()
      exp = codeAndExp[1].lstrip() if len(codeAndExp) > 1 else ""
      realCodeObj = removeBackTick(code)
      pres.append(keepSuffix(lastLineCode, realCodeObj))
      pexp.append(exp)
  return pres, pexp


import pandas as pd
import json
from typing import List, Optional, Union

from . import utils as utilsXX

def getShape(dfName: pd.DataFrame): 
    print(dfName.shape)

def _getColumns(dfName: pd.DataFrame):
    # type
    typeDF = dfName.dtypes.reset_index().rename(columns={"index": "colName", 0: "dtype"})
    # null count
    nullCountDf = dfName.isnull().sum().reset_index().rename(columns={"index": "colName", 0: "nullCount"})
    # cardinality, min, max
    desDf = dfName.describe(include="all").T.reset_index()
    desDfColumns = desDf.columns.tolist()
    if "unique" not in desDfColumns:
        desDf["unique"] = pd.Series()
    if "min" not in desDfColumns:
        desDf["min"] = pd.Series()
    if "max" not in desDfColumns:
        desDf["max"] = pd.Series()
    desDf = desDf[["index", "unique", "min", "max"]].rename(columns={"index": "colName", "unique": "cardinality", "min": "minValue", "max": "maxValue"})
    # sortedness
    sortDf = dfName.apply(lambda x: "asc" if x.is_monotonic_increasing else "desc" if x.is_monotonic_decreasing else "no").to_frame().reset_index().rename(columns={"index": "colName", 0: "sortedness"})
    res = typeDF.merge(nullCountDf, on="colName").merge(desDf, on="colName").merge(sortDf, on="colName")
    return res

def getColumns(dfName: pd.DataFrame):
    res = _getColumns(dfName)
    print(res.to_json(orient="records", default_handler=str))

def getHeadTail(dfName: pd.DataFrame, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    # res = pd.concat([dfName.head(n), dfName.tail(n)])
    res = dfName.head(n)
    print(res.to_json(orient="records", default_handler=str))

def getCateValueCounts(dfName: pd.DataFrame, colName: str, limit: int = 1000):
    vcnt_ser = dfName[colName].value_counts()
    vcnt_frame = vcnt_ser.to_frame().reset_index().head(limit)
    vcnt_frame = vcnt_frame.rename(columns={colName: "value"})
    print(vcnt_frame.to_json(orient="records", default_handler=str))

def getRowSampleOfOneDf(dfName: pd.DataFrame, condJsonStr: str, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    condDict: dict = json.loads(condJsonStr)
    for col, cond in condDict.items():
        if cond["cateSelect"] is not None:
            dfName = dfName[dfName[col].isin(cond["cateSelect"])]
        elif cond["rangeSelect"] is not None:
            dfName = utilsXX.selectRowsWithinRange(dfName, col, cond["rangeSelect"])
    print(dfName.head(n).to_json(orient="records", default_handler=str))

def getNumColumnSample(dfName: pd.DataFrame, colName: str, n: int = 1000):
    cleanedCol = dfName[colName].dropna()
    res = {
        "sample": cleanedCol.sample(min(len(cleanedCol), n)).to_list()
    }
    print(json.dumps(res, default=str))

def getNumHistogram(dfName: pd.DataFrame, colName: str):
    data = _getNumHistogram(dfName, colName)
    print(json.dumps(data, default=str))

def _getNumHistogram(dfName: pd.DataFrame, colName: str):    
    # If all nulls
    if dfName[colName].isna().sum() == len(dfName[colName]):
        data = {
            "histogram": []
        }
        return data
    
    minV = None
    chartData = None
    # TODO: We treat time series as a special case for numerical data
    if str(dfName[colName].dtype) in ['datetime64', 'datetime64[ns]']:
        minV = (dfName[colName].astype("int64")//1e9).min()
        chartData = utilsXX.getTempBinnedData(dfName[colName])
    else:
        minV = dfName[colName].min()
        chartData = utilsXX.getQuantBinnedData(dfName[colName])
    chartData = utilsXX.convertBinned(chartData, minV)
    data = {
        "histogram": chartData
    }
    return data

def exeFillna(dfName: pd.DataFrame, colName: str, fillValue: Union[str, int, float], n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    s = dfName[colName].fillna(fillValue)
    sf = s.to_frame()
    meta = _getColumns(sf)
    data = {
        "newList": s.head(n).to_list(),
        "meta": meta.to_json(orient="records", default_handler=str)
    }
    print(json.dumps(data, default=str))
    
def exeColumnRename(dfName: pd.DataFrame, oldColName: str, newColName: str, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    s = dfName.rename(columns={oldColName: newColName})
    meta = _getColumns(s)
    data = {
        "newList": s[newColName].head(n).to_list(),
        "meta": meta.to_json(orient="records", default_handler=str)
    }
    print(json.dumps(data, default=str))
   

def exeColumnAdd(dfName: pd.DataFrame, colName: str, operator: str, addedColumnsStr: str, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):    
    
    addedColumns: List[str] = json.loads(addedColumnsStr)

    s = dfName[addedColumns[0]]

    for i in range(1, len(addedColumns)):
        if operator == "+":
            s = s + dfName[addedColumns[i]]
        elif operator == "/":
            s = s / dfName[addedColumns[i]]

    if operator == "+":
        if colName in dfName.columns:
            s + dfName[colName]

    sf = s.to_frame()
    meta = _getColumns(sf)
    data = {
        "newList": s.head(n).to_list(),
        "meta": meta.to_json(orient="records", default_handler=str)
    }
    print(json.dumps(data, default=str))
  
def exeStrReplace(series: List[Optional[str]], old: str, newStr: str):
    # convert to pd series
    seriesPd = pd.Series(series)
    # replace
    seriesPd = seriesPd.str.replace(old, newStr)
    # convert to list
    print(seriesPd.to_json(orient="records", default_handler=str))

def exeStrReplaceMeta(dfName: pd.DataFrame, colName: str, old: str, newStr: str):
    # deep copy dfName
    res = dfName.copy()[[colName]]
    res[colName] = res[colName].str.replace(old, newStr)
    meta = _getColumns(res)
    print(meta.to_json(orient="records", default_handler=str))

def exeTableFilter(df: pd.DataFrame, dfName: str, condStr: str, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    condition = eval(condStr, {dfName: df})
    res = df[condition]
    columns = _getColumns(res)
    columnNameList = res.columns.tolist()
    histograms = {}
    for col in columnNameList:
        if utilsXX.isNumColumn(str(res[col].dtype)):
            histograms[col] = _getNumHistogram(res, col)["histogram"]

    allJson = {
        "rows": res.head(n).to_json(orient="records", default_handler=str),
        "columns": columns.to_json(orient="records", default_handler=str),
        "numRows": res.shape[0],
        "numCols": res.shape[1],
        "columnNameList": columnNameList,
        "histograms": histograms
    }
    print(json.dumps(allJson, default=str))

def exeTableSort(df: pd.DataFrame, sortColumnsStr: str, ascendingStr: str, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    sortColumns: List[str] = json.loads(sortColumnsStr)
    ascending: List[bool] = json.loads(ascendingStr)
    res = df.sort_values(by=sortColumns, ascending=ascending)
    columns = _getColumns(res)
    columnNameList = res.columns.tolist()
    histograms = {}
    for col in columnNameList:
        if utilsXX.isNumColumn(str(res[col].dtype)):
            histograms[col] = _getNumHistogram(res, col)["histogram"]
    
    allJson = {
        "rows": res.head(n).to_json(orient="records", default_handler=str),
        "columns": columns.to_json(orient="records", default_handler=str),
        "numRows": res.shape[0],
        "numCols": res.shape[1],
        "columnNameList": columnNameList,
        "histograms": histograms
    }
    print(json.dumps(allJson, default=str))

def exeGroupby(df: pd.DataFrame, groupbyColumnsStr: str, selectColumnsStr: str, aggFunc: str, n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    
    groupbyColumns: List[str] = json.loads(groupbyColumnsStr)
    selectColumns: List[str] = json.loads(selectColumnsStr)
    if(selectColumns == None):
        res = df.groupby(groupbyColumns).agg(aggFunc) 
    else:
        res = df.groupby(groupbyColumns)[selectColumns].agg(aggFunc)

    res = res.reset_index()
    columns = _getColumns(res)
    columnNameList = res.columns.tolist()
    histograms = {}
    for col in columnNameList:
        if utilsXX.isNumColumn(str(res[col].dtype)):
            histograms[col] = _getNumHistogram(res, col)["histogram"]
    
    allJson = {
        "rows": res.head(n).to_json(orient="records", default_handler=str),
        "columns": columns.to_json(orient="records", default_handler=str),
        "numRows": res.shape[0],
        "numCols": res.shape[1],
        "columnNameList": columnNameList,
        "histograms": histograms
    }
    print(json.dumps(allJson, default=str))

def exeTableMerge(leftDf: pd.DataFrame, rightDf: pd.DataFrame, leftCol: Optional[str], rightCol: Optional[str], how: Optional[str], n: int = utilsXX.DEFAULT_ROW_SAMPLE_SIZE):
    res = leftDf
    if how is not None:
        res = pd.merge(leftDf, rightDf, left_on=leftCol, right_on=rightCol, how=how)
    else:
        res = pd.merge(leftDf, rightDf, left_on=leftCol, right_on=rightCol)
    columns = _getColumns(res)
    columnNameList = res.columns.tolist()
    histograms = {}
    for col in columnNameList:
        if utilsXX.isNumColumn(str(res[col].dtype)):
            histograms[col] = _getNumHistogram(res, col)["histogram"]
    allJson = {
        "rows": res.head(n).to_json(orient="records", default_handler=str),
        "columns": columns.to_json(orient="records", default_handler=str),
        "numRows": res.shape[0],
        "numCols": res.shape[1],
        "columnNameList": columnNameList,
        "histograms": histograms
    }
    print(json.dumps(allJson, default=str))
    

# def getAllCateColumnFmt(dfName: pd.DataFrame):
#     res = dfName.select_dtypes(include=["object", "category"])
#     cateColList = list(res.columns)
#     res = []
#     for col in cateColList:
#         res.append({
#             "colName": col,
#             "isalpha": dfName[col].str.isalpha().sum(),
#             "isnumeric": dfName[col].str.isnumeric().sum(),
#             "isalnum": dfName[col].str.isalnum().sum(),
#             "isdigit": dfName[col].str.isdigit().sum(),
#             "isdecimal": dfName[col].str.isdecimal().sum(),
#             "isspace": dfName[col].str.isspace().sum(),
#             "islower": dfName[col].str.islower().sum(),
#             "isupper": dfName[col].str.isupper().sum(),
#             "istitle": dfName[col].str.istitle().sum()
#         })
#     print(pd.DataFrame(res).to_json(orient="records", default_handler=str))


import re
from typeguard import typechecked

from .datatypes import List, Dict, Any, Union, Optional, ColumnInfo
from .constant import AST_POS, COMPLETION_ITEM_KIND, SPECIAL_CASE, SptMethodName
from .debugger import debugger

@typechecked
def getCompletionItemTypeFromAstPos(astPos: int) -> str:
  kind = COMPLETION_ITEM_KIND.get(astPos)
  if kind is None:
    # debugger.warning(f"[getCompletionItemTypeFromAstPos] Invalid astPos: {astPos}")
    return ""
  else:
    return kind

@typechecked
def code2DTo0D(code: List[List[str]]) -> str:
  def fn(x: List[str]) -> str:
    return "\n".join(x)
  
  code1D = list(map(fn, code))
  return "\n\n".join(code1D)

@typechecked
def removeQuotes(s: str) -> str:
  if s.startswith("'''") and s.endswith("'''"):
    return s[3:-3] 
  elif s.startswith('"""') and s.endswith('"""'):
    return s[3:-3] 
  elif s.startswith('"') and s.endswith('"'):
    return s[1:-1]
  elif s.startswith("'") and s.endswith("'"):
    return s[1:-1]
  else:
    return s
  
@typechecked
def findAllVarLastOccur(code: str, varNameList: List[str]) -> Dict[str, int]:
  lastOccurDict: Dict[str, int] = {}
  for varName in varNameList:
    lastOccur = code.rfind(varName)
    if lastOccur != -1:
      lastOccurDict[varName] = lastOccur
  return lastOccurDict

@typechecked
def sortDfByOccur(lastOccurDict: Dict[str, int], cxt_dfName: Optional[str] = None) -> List[str]:
  sortedVarList = sorted(lastOccurDict.items(), key=lambda x: x[1], reverse=True)
  sortedVarNameList = [x[0] for x in sortedVarList]
  if (cxt_dfName is not None) and (cxt_dfName in sortedVarNameList):
    sortedVarNameList.remove(cxt_dfName)
    sortedVarNameList.append(cxt_dfName)
  return sortedVarNameList

@typechecked
def getColInfoByName(colArr: List[ColumnInfo], colName: str) -> Optional[ColumnInfo]:
  col_info = None
  for col in colArr:
    if col["colName"] == colName:
      col_info = col
      break
  return col_info

@typechecked
def fullWordMatch(word: str, sentence: str) -> Optional[re.Match[str]]:
  pattern = r"\b" + re.escape(word) + r"\b"
  return re.search(pattern, sentence)


@typechecked
def needAddPrefix(special_case: int, is_trivial: bool, need_obj: int, method_name: Optional[str], cell_value: Optional[str]) -> bool:
  # cases that no need to add prefix
  case1 = (special_case == SPECIAL_CASE.NONE) and is_trivial and (need_obj == AST_POS.CELL_VALUE) and (method_name == SptMethodName.VALUE_FILTER) and (cell_value is not None)
  return (not case1)


def formatNumber(num: Union[int, float]) -> Union[int, float]:
  if isinstance(num, int):
    return num
  elif isinstance(num, float):
    return round(num, 2)
  else:
    # debugger.warning(f"[format_number] Invalid input type: {type(num)}")
    return num

@typechecked
def isCateColumn(dtype: str) -> bool:
  return dtype in ["object", "string", "str", "category"]

@typechecked
def isTimeColumn(dtype: str) -> bool:
  return dtype in ["datetime64", "datetime64[ns]"]

@typechecked
def isNumColumn(dtype: str) -> bool:
  INT_TYPE = ['int', 'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64', 'Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32', 'UInt64']
  FLOAT_TYPE = ['float', 'float_', 'float16', 'float32', 'float64']
  NUM_TYPE = [*INT_TYPE, *FLOAT_TYPE]
  return dtype in NUM_TYPE or isTimeColumn(dtype)

@typechecked
def logInFiles(msg: str):
  with open("log.txt", "a", encoding="utf-8") as f:
    f.write(msg + "\n\n")


import logging
from typeguard import typechecked

class CustomFormatter(logging.Formatter):
  grey = "\x1b[38;20m"
  yellow = "\x1b[33;20m"
  red = "\x1b[31;20m"
  bold_red = "\x1b[31;1m"
  reset = "\x1b[0m"
  # str_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
  str_format = "[%(filename)s:%(lineno)d] %(levelname)s: %(message)s "

  FORMATS = {
    logging.DEBUG: grey + str_format + reset,
    logging.INFO: grey + str_format + reset,
    logging.WARNING: yellow + str_format + reset,
    logging.ERROR: red + str_format + reset,
    logging.CRITICAL: bold_red + str_format + reset
  }

  @typechecked
  def format(self, record):
    log_fmt = self.FORMATS.get(record.levelno)
    formatter = logging.Formatter(log_fmt)
    return formatter.format(record)
  
@typechecked
def initLogger():
  # create logger with 'spam_application'
  logger = logging.getLogger("Xavier_Prompt")
  logger.setLevel(logging.DEBUG)

  # create console handler with a higher log level
  ch = logging.StreamHandler()
  ch.setLevel(logging.DEBUG)
  ch.setFormatter(CustomFormatter())
  logger.addHandler(ch)
  return logger

debugger = initLogger()


from typeguard import typechecked
from ..constant import PROMPT_MARKER

@typechecked
def codePreamble() -> str:
  return f"""Suppose a user has typed some Python code below. Here is the {PROMPT_MARKER.CODE_CONTEXT} before the input cursor {PROMPT_MARKER.CURSOR}:"""

@typechecked
def wrapPythonCode(code: str) -> str:
  return f"""```python
{code}{PROMPT_MARKER.CURSOR}
```"""

@typechecked
def getCodeInfo(pCode: str) -> str:
  prompt = codePreamble() + "\n\n"
  prompt += f"{wrapPythonCode(pCode)}\n"
  return prompt


import copy
from typeguard import typechecked

from ..lexAnalysis.pyParse import getPandasAlias, matchRulesPDFuncSig, matchRulesGlobal, matchRulesDfFuncSig, matchRulesSeFuncSig, matchRulesSeStrFuncSig, getSTFromMatchDetail, isComment
from ..lexAnalysis import pyParse as pyParseX
from .. import datatypes as dtX
from ..prompt import createPrompt as cpX
from .. import utils as utilsX
from ..lexAnalysis import fixJedi as fixJediX

from ..server.myAIClient import myAIClient
from ..lexAnalysis.pyParse import get_tokens_from_code
from ..constant import AST_POS, SpecialTokens, SPECIAL_CASE, SptMethodName, DF_INFO_TYPE, PROMPT_MARKER
from ..datatypes import List, Tuple, Optional, PartialCodeInfo, CompletionItem, JupyterlabToken, TokenInfo
from ..trivial.trivialComplete import trivialComplete
from ..trivial import specialCaseComplete as sccX
from ..debugger import debugger

@typechecked
def cacheCodeMatch(pTokens: List[TokenInfo], cacheTokens: List[TokenInfo]) -> bool:
  if len(pTokens) != len(cacheTokens):
    return False
  for i in range(len(pTokens)):
    case1 = (pTokens[i]["exact_type"] != cacheTokens[i]["exact_type"])
    case2 = (pTokens[i]["string"] != cacheTokens[i]["string"])
    case3 = (pTokens[i]["type"] != cacheTokens[i]["type"])
    if case1 or case2 or case3:
      return False
  return True

@typechecked
def try_complete(client: myAIClient, previousCode2D: List[List[str]], token: JupyterlabToken, tableLvInfo: dtX.TableLevelInfoT, colLvInfo: dtX.ColLevelInfoT, rowLvInfo: dtX.RowLevelInfoT) -> Tuple[List[CompletionItem], PartialCodeInfo]:
  analyze_resp = try_analyze_code(previousCode2D, tableLvInfo)

  # debugger.info(f"analyze_resp: {analyze_resp}")
  # return []

  previousCode = utilsX.code2DTo0D(previousCode2D)
  pTokens: List[TokenInfo] = get_tokens_from_code(previousCode, True)
  token_list: List[CompletionItem] = []
  cacheUsed: bool = False
  need_obj = analyze_resp["need_obj"]
  special_case = analyze_resp["special_case"]
  is_trivial = analyze_resp["is_trivial"]

  if not cacheUsed:
#     prompt = "Suppose you are a code autocompletion tool, and a user has typed some Python code below. Please complete the code.\n\n"
#     prompt += shareX.wrapPythonCode(previousCode) + "\n\n"
#     prompt += f"""please: (1) think of the possible codes (e.g. data transformation operations applied to the datasets) the user wants to write; (2) rank these codes based on the correlation between the code and "{PROMPT_MARKER.CODE_CONTEXT}"; (3) provide the autocompletion prompt and add to {PROMPT_MARKER.CURSOR} of the code, {fmtCtrlX.fmtCtrl_noExplanation(False)}. The output format is like:
# 1. <rest part of code line 1>
# 2. <rest part of code line 2>
# ...
# {fmtCtrlX.fmtCtrl_additionalNotes("codeLine", False, fmtCtrlX.forbiddenFunc())}"""
#     token_list = cpX.gptComplete_Exe(client, token, prompt, previousCode, analyze_resp["need_obj"])


    if special_case != SPECIAL_CASE.NONE:
      # special case handling
      token_list = sccX.specialCaseComplete(client, token, tableLvInfo, rowLvInfo, previousCode2D, analyze_resp)
    elif is_trivial:
      # 2. some trivial cases can be handled directly
      token_list = trivialComplete(need_obj, token, previousCode, analyze_resp, tableLvInfo)
    else:
      # 3. GPT-4 completion for more complex cases
      prompt: Optional[str] = None
      filteredTLI, filteredCLI, filteredRLI = cpX.filterDataContext(analyze_resp, tableLvInfo, colLvInfo, rowLvInfo)
      if (filteredTLI is not None) or (filteredCLI is not None) or (filteredRLI is not None):
        prompt = cpX.gptCompleteForSC(previousCode, analyze_resp["method_name"], analyze_resp["need_obj"], filteredTLI, filteredCLI, filteredRLI)
      if prompt is not None:
        token_list = cpX.gptComplete_Exe(client, token, prompt, previousCode, analyze_resp["need_obj"])


  # Add the token value as the prefix to the completion items
  if utilsX.needAddPrefix(special_case, is_trivial, need_obj, analyze_resp["method_name"], analyze_resp["cell_value"]):
    for i in range(len(token_list)):
      token_list[i]["value"] = token["value"] + token_list[i]["value"]

  return token_list, analyze_resp

@typechecked
def try_analyze_code(code: List[List[str]], tableLvInfo: dtX.TableLevelInfoT) -> PartialCodeInfo:
  pandasAlias = getPandasAlias(code)
  lastline = code[-1][-1]
  lastline_tokens = get_tokens_from_code(lastline, True)
  secondLastline = code[-1][-2] if len(code[-1]) >= 2 else None
  secondLastline_tokens = get_tokens_from_code(secondLastline, True) if secondLastline is not None else None
  res: PartialCodeInfo = {
    "method_name": None,
    "need_obj": AST_POS.INVALID,
    "df_name": None,
    "df_name_list": None,
    "col_name": None,
    "col_name_list": None,
    "col_name_quote": None,
    "op_name": None,
    "cell_value": None,
    "col_idx_list": None,
    "df_info_type": "",
    "special_case": SPECIAL_CASE.NONE,
    "is_trivial": False,
  }
  
  sigs = fixJediX.get_signature_fixjedi(lastline, pandasAlias)
  scNo = pyParseX.specialCaseDetect(lastline_tokens, pandasAlias, tableLvInfo)
  final_rule, match_detail = matchRulesGlobal(lastline_tokens, pandasAlias, tableLvInfo)
  need_obj_maybe = final_rule.get("need_obj", AST_POS.INVALID)
  df_info_type_maybe = final_rule.get("df_info_type", "")
  method_name_maybe = final_rule.get("full_name", "")
  is_trivial_maybe = final_rule.get("is_trivial", False)


  if len(sigs) == 0:
    if isComment(lastline_tokens):
      # Comment to comment
      res["is_trivial"] = False
      res["need_obj"] = AST_POS.COMMENT
      res["df_info_type"] = DF_INFO_TYPE.USER_SPECIFIED
      res["method_name"] = SptMethodName.COMMENT
    elif lastline == "" and secondLastline_tokens and isComment(secondLastline_tokens):
      # Comment to code
      res["is_trivial"] = False
      res["need_obj"] = AST_POS.CODE_LINE
      res["df_info_type"] = DF_INFO_TYPE.USER_SPECIFIED
      res["method_name"] = SptMethodName.CODE_LINE
    elif scNo == SPECIAL_CASE.DF_NAME_PREFIX:
      res["method_name"] = SptMethodName.PREFIX_DF_NAME
      res["need_obj"] = AST_POS.CODE_LINE
      res["df_name_list"] = sccX.getItemsByPrefix(lastline_tokens[-1]["string"], list(tableLvInfo.keys()))
      res["special_case"] = scNo
    elif scNo == SPECIAL_CASE.DF_SELECT:
      res["method_name"] = SptMethodName.DF_SELECT
      res["need_obj"] = AST_POS.CODE_LINE
      res["df_name"] = lastline_tokens[-2]["string"]
      res["special_case"] = scNo
    elif scNo == SPECIAL_CASE.DF_SELECT_COL_1:
      res["method_name"] = SptMethodName.PREFIX_COL_NAME
      res["need_obj"] = AST_POS.CODE_LINE
      res["col_name_quote"] = lastline_tokens[-1]["string"]
      res["df_name"] = lastline_tokens[-3]["string"]
      res["col_name_list"] = tableLvInfo[res["df_name"]]["columnNameList"]
      res["special_case"] = scNo
    elif scNo == SPECIAL_CASE.DF_SELECT_COL_2:
      res["method_name"] = SptMethodName.PREFIX_COL_NAME
      res["need_obj"] = AST_POS.CODE_LINE
      res["col_name_quote"] = lastline_tokens[-2]["string"]
      res["df_name"] = lastline_tokens[-4]["string"]
      res["col_name_list"] = sccX.getItemsByPrefix(lastline_tokens[-1]["string"], tableLvInfo[res["df_name"]]["columnNameList"])
      res["special_case"] = scNo
    elif scNo == SPECIAL_CASE.DF_SELECT_COL_3:
      res["method_name"] = SptMethodName.PREFIX_COL_NAME
      res["need_obj"] = AST_POS.CODE_LINE
      res["col_name_quote"] = lastline_tokens[-2]["string"]
      res["df_name"] = None
      res["col_name_list"] = sccX.getItemsByPrefix(lastline_tokens[-1]["string"], sccX.getAllColumnNames(tableLvInfo))
      res["special_case"] = scNo
    elif method_name_maybe == SptMethodName.DF_COL:
      res["method_name"] = method_name_maybe
      res["need_obj"] = need_obj_maybe
      res["df_name"] = getSTFromMatchDetail(match_detail, SpecialTokens.DF)
      res["col_name"] = utilsX.removeQuotes(getSTFromMatchDetail(match_detail, SpecialTokens.COLIDX))
    elif method_name_maybe == SptMethodName.VALUE_FILTER:
      maybeDfNames = getSTFromMatchDetail(match_detail, SpecialTokens.DF, "all")
      if len(list(set(maybeDfNames))) == 1:
        res["method_name"] = method_name_maybe
        res["need_obj"] = need_obj_maybe
        res["df_name"] = maybeDfNames[0]
        res["col_name"] = utilsX.removeQuotes(getSTFromMatchDetail(match_detail, SpecialTokens.COLIDX))
    elif method_name_maybe == SptMethodName.COLUMNS_SELECT and is_trivial_maybe:
      res["method_name"] = method_name_maybe
      res["need_obj"] = need_obj_maybe
      res["df_name"] = getSTFromMatchDetail(match_detail, SpecialTokens.DF)
      res["col_name_list"] = tableLvInfo[res["df_name"]]["columnNameList"]
      res["is_trivial"] = is_trivial_maybe
    elif method_name_maybe == SptMethodName.GROUPBY_COLSELECT and is_trivial_maybe:
      res["method_name"] = method_name_maybe
      res["need_obj"] = need_obj_maybe
      res["df_name"] = getSTFromMatchDetail(match_detail, SpecialTokens.DF)
      res["col_name_list"] = tableLvInfo[res["df_name"]]["columnNameList"]
      res["is_trivial"] = is_trivial_maybe
    else:
      res["special_case"] = scNo
      res["need_obj"] = need_obj_maybe
      res["df_info_type"] = df_info_type_maybe
      res["method_name"] = method_name_maybe
      res["is_trivial"] = is_trivial_maybe
      res["df_name"] = getSTFromMatchDetail(match_detail, SpecialTokens.DF)
      res["col_name"] = getSTFromMatchDetail(match_detail, SpecialTokens.COLIDX)
      res["op_name"] = getSTFromMatchDetail(match_detail, SpecialTokens.OP)
      res["cell_value"] = getSTFromMatchDetail(match_detail, SpecialTokens.CELL_VALUE)
      res["col_idx_list"] = getSTFromMatchDetail(match_detail, SpecialTokens.COLIDX_LIST)
      if res["col_name"] is not None:
        res["col_name"] = utilsX.removeQuotes(res["col_name"])
  else:
    sig = sigs[0] # @TODO: Only consider the first signature
    if scNo == SPECIAL_CASE.DF_SELECT_COL_3:
      res["method_name"] = SptMethodName.PREFIX_COL_NAME
      res["need_obj"] = AST_POS.CODE_LINE
      res["col_name_quote"] = lastline_tokens[-2]["string"]
      res["df_name"] = None
      res["col_name_list"] = sccX.getItemsByPrefix(lastline_tokens[-1]["string"], sccX.getAllColumnNames(tableLvInfo))
      res["special_case"] = scNo
    elif sig["type"] == "function" and sig["module_name"].startswith("pandas.core.reshape"):
      final_rule = matchRulesPDFuncSig(lastline_tokens, pandasAlias, sig["name"])
      res["need_obj"] = final_rule.get("need_obj", AST_POS.INVALID)
      res["df_info_type"] = final_rule.get("df_info_type", "")
      res["method_name"] = final_rule.get("full_name", "")
      res["is_trivial"] = final_rule.get("is_trivial", False)
    elif sig["type"] == "function" and sig["module_name"].startswith("pandas.core.frame"):
      final_rule, dfName = matchRulesDfFuncSig(lastline_tokens, tableLvInfo, sig["name"])
      res["need_obj"] = final_rule.get("need_obj", AST_POS.INVALID)
      res["df_info_type"] = final_rule.get("df_info_type", "")
      res["method_name"] = final_rule.get("full_name", "")
      res["is_trivial"] = final_rule.get("is_trivial", False)
      res["df_name"] = dfName
      if res["method_name"] == "" and dfName != "" and lastline_tokens[-1]["string"] in ["'", '"']:
        res["need_obj"] == AST_POS.CODE_LINE
        res["method_name"] = SptMethodName.PREFIX_COL_NAME
        res["col_name_list"] = tableLvInfo[dfName]["columnNameList"]
        res["col_name_quote"] = lastline_tokens[-1]["string"]
        res["special_case"] = SPECIAL_CASE.DF_SELECT_COL_3
    elif sig["type"] == "function" and sig["module_name"].startswith("pandas.core.series"):
      final_rule, dfName, colName = matchRulesSeFuncSig(lastline_tokens, pandasAlias, tableLvInfo, sig["name"])
      res["need_obj"] = final_rule.get("need_obj", AST_POS.INVALID)
      res["df_info_type"] = final_rule.get("df_info_type", "")
      res["method_name"] = final_rule.get("full_name", "")
      res["is_trivial"] = final_rule.get("is_trivial", False)
      res["df_name"] = dfName
      res["col_name"] = colName
    elif sig["type"] == "function" and sig["module_name"].startswith("pandas.core.strings.accessor"):
      final_rule, dfName, colName = matchRulesSeStrFuncSig(lastline_tokens, pandasAlias, tableLvInfo, sig["name"])
      res["need_obj"] = final_rule.get("need_obj", AST_POS.INVALID)
      res["df_info_type"] = final_rule.get("df_info_type", "")
      res["method_name"] = final_rule.get("full_name", "")
      res["is_trivial"] = final_rule.get("is_trivial", False)
      res["df_name"] = dfName
      res["col_name"] = colName

  # debugger.info(f"[try_analyze_code] res: {res} \nsig: {sigs}")
  return res




from typeguard import typechecked
from typing import Union, List, Optional

from .. import datatypes as dtX
from ..constant import PROMPT_MARKER
from .. import utils as utilsX

@typechecked
def dataInfoPreamble() -> str:
  return f"""Here is {PROMPT_MARKER.DATA_CONTEXT} in this code script:"""

@typechecked
def multiTableLvInfoPrompt(tableLvInfo: dtX.TableLevelInfoT) -> str:
  p: str = ""
  for tName, item in tableLvInfo.items():
    p += f"""{tableLevelInfoPrompt(tName, item["numRows"], item["numCols"], item["columnNameList"])}\n"""
  return p

@typechecked
def tableLevelInfoPrompt(dfName: str, num_rows: Optional[int] = None, num_cols: Optional[int] = None, colNames: Optional[List[str]] = None) -> str:
  prompt = f"""DataFrame: {dfName}"""
  if num_rows is not None and num_cols is not None:
    prompt += f""" ({num_rows} rows x {num_cols} columns)"""
  if colNames is not None:
    prompt += f"""\nColumns: {", ".join(colNames)}"""
  return prompt

@typechecked
def multiRowLvInfoPrompt(rowLvInfo: dtX.RowLevelInfoT) -> str:
  p: str = ""
  for tName, item in rowLvInfo.items():
    p += f"""{rowLevelInfoPrompt(tName, item["columnNameList"], item["sampleRows"])}\n"""
  return p

@typechecked
def rowLevelInfoPrompt(dfName: str, colNameList: List[str], sampleRows: Optional[List[dtX.DFRowT]] = None) -> str:
  prompt: str = ""
  if (sampleRows is not None) and (len(colNameList) > 0) and (len(sampleRows) > 0):
    prompt += f"""DataFrame: {dfName}\n"""
    prompt += f"""Sample rows:\n"""
    prompt += getTableHeaderMDFmt(colNameList) + "\n"
    for row in sampleRows:
      prompt += getTableBodyMDFmt(colNameList, [row])
  return prompt

@typechecked
def multiColLvInfoPrompt(colLvInfo: dtX.ColLevelInfoT, tableLvInfo: Optional[dtX.TableLevelInfoT]) -> str:
  p: str = ""
  for tName, titem in colLvInfo.items():
    p += f"""DataFrame: {tName}\n"""
    tLvInfo = tableLvInfo[tName] if tableLvInfo is not None else None
    for cName, citem in titem.items():
      p += colLevelInfoPrompt("\t", cName, citem, tLvInfo)
  return p

@typechecked
def colLevelInfoPrompt(indent: str, colName: str, citem: dtX._ColLevelInfoT, tLvInfo: Union[None, dtX._TableLevelInfoT]) -> str:
  prompt = f"""{indent}Column: {colName}"""
  if citem["dtype"] is not None:
    prompt += f""" (Data type: {citem["dtype"]})\n"""
  else:
    prompt += "\n"
  if citem["nullCount"] is not None and citem["nullCount"] > 0:
    prompt += f"""{indent}{indent}null_count: {citem["nullCount"]}\n"""
  if citem["sortedness"] is not None:
    if citem["sortedness"] == "asc":
      prompt += f"""{indent}{indent}The values are sorted in ascending order.\n"""
    elif citem["sortedness"] == "desc":
      prompt += f"""{indent}{indent}The values are sorted in descending order.\n"""
    else:
      prompt += f"""{indent}{indent}The values are not sorted.\n"""

  if utilsX.isNumColumn(citem["dtype"]):
    if citem["minValue"] is not None and citem["maxValue"] is not None:
      prompt += f"""{indent}{indent}min: {citem["minValue"]}\n"""
      prompt += f"""{indent}{indent}max: {citem["maxValue"]}\n"""
    if citem["sample"] is not None and len(citem["sample"]) > 0:
      prompt += f"""{indent}{indent}Sample values (format: value1, value2, ...): {", ".join([str(v) for v in citem["sample"]])}\n"""
  else:
    if citem["cardinality"] is not None:
      prompt += f"""{indent}{indent}cardinality: {citem["cardinality"]}"""
      numRows = tLvInfo["numRows"] if tLvInfo is not None else None
      percent = citem["cardinality"] / numRows * 100 if numRows is not None and numRows > 0 else None
      if numRows is not None and numRows == citem["cardinality"]:
        prompt += " (all unique values)"
      elif percent is not None:
        prompt += f""" ({percent:.2f}% of the total rows)"""
      prompt += "\n"
    # if citem["uniqueValues"] is not None and citem["uvCounts"] is not None:
    #   if len(citem["uniqueValues"]) > 0:
    #     uniVals = citem["uniqueValues"][:100] # TODO: truncate the unique values
    #     uniCnts = citem["uvCounts"][:100] # TODO: truncate the unique values
    #     prompt += f"""{indent}{indent}Unique values (format: "value1"(count1), "value2"(count2), ...): """
    #     prompt += ", ".join([f"\"{uv}\"({cnt})" for uv, cnt in zip(uniVals, uniCnts)]) + "\n"
    # elif citem["uniqueValues"] is not None:
    if citem["uniqueValues"] is not None:
      if len(citem["uniqueValues"]) > 0:
        uniVals = citem["uniqueValues"]
        addPrompt = f"""{indent}{indent}Unique values (format: "value1", "value2", ...): {", ".join(['"' + str(uv) + '"' for uv in uniVals])}\n"""
        addPrompt = addPrompt[:1500] # TODO: truncate the prompt
        prompt += addPrompt
  return prompt

@typechecked
def getTableHeaderMDFmt(colNames: List[str]) -> str:
  firstLine = f"""| {" | ".join(colNames)} |"""
  secondLine = f"""| {" | ".join([":---:"] * len(colNames))} |"""
  return firstLine + "\n" + secondLine

@typechecked
def getTableBodyMDFmt(colNames: List[str], rows: List[dtX.DFRowT]) -> str:
  prompt: str = ""
  for row in rows:
    prompt += f"""| {" | ".join([str(row[col]) for col in colNames])} |\n"""
  return prompt



from typeguard import typechecked

from ..utils import getCompletionItemTypeFromAstPos, removeQuotes, findAllVarLastOccur, sortDfByOccur, getColInfoByName, formatNumber
from ..constant import AST_POS, SptMethodName
from ..datatypes import List, Tuple, Union, PartialCodeInfo, CompletionItem, JupyterlabToken, NumericalStatsT
from ..docCtrl import genCtrlDocumentation
from ..debugger import debugger

from .. import datatypes as dtX
from ..prompt import parseGPTOutput as parseGPTX

@typechecked
def dfVarComplete(need_obj: int, token: JupyterlabToken, previousCode: str, analyze_resp: PartialCodeInfo, allDfInfo) -> List[CompletionItem]:
  all_df_names = list(allDfInfo.keys())
  dfOcc = findAllVarLastOccur(previousCode, all_df_names)
  ordered_dfName = sortDfByOccur(dfOcc, analyze_resp["df_name"])
  return list(map(lambda x: CompletionItem(
    value=x,
    offset=token["offset"],
    type=getCompletionItemTypeFromAstPos(need_obj),
    explanation=None
  ), ordered_dfName))

@typechecked
def paramComplete(need_obj: int, token: JupyterlabToken, previousCode: str, analyze_resp: PartialCodeInfo, allDfInfo) -> List[CompletionItem]:
  if analyze_resp["method_name"] == SptMethodName.COL_ASTYPE:
    typearr = ["int", "float", "str", "bool"]
    return list(map(lambda x: CompletionItem(
      value=x,
      offset=token["offset"],
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=None
    ), typearr))
  return []

@typechecked
def defaultNumCompleteByStats(need_obj: int, token_offset: int, numStats: Union[None, NumericalStatsT], df_info_type: str, known_df_name: str, known_col_name: str, allDfInfo) -> List[CompletionItem]:
  if numStats is None:
    # debugger.warning(f"[trivialComplete] cell_value for value_filter should have valid numStats, but got {numStats}")
    return []

  q50 = formatNumber(numStats["q50"])
  q25 = formatNumber(numStats["q25"])
  q75 = formatNumber(numStats["q75"])
  min_val = formatNumber(numStats["min"])
  max_val = formatNumber(numStats["max"])
  mean_val = formatNumber(numStats["mean"])

  return [
    CompletionItem(
      value=str(q50),
      offset=token_offset,
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=genCtrlDocumentation(df_info_type, str(q50), allDfInfo, known_df_name, known_col_name, str(q50), None)
    ),
    CompletionItem(
      value=str(q25),
      offset=token_offset,
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=genCtrlDocumentation(df_info_type, str(q25), allDfInfo, known_df_name, known_col_name, str(q25), None)
    ),
    CompletionItem(
      value=str(q75),
      offset=token_offset,
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=genCtrlDocumentation(df_info_type, str(q75), allDfInfo, known_df_name, known_col_name, str(q75), None)
    ),
    CompletionItem(
      value=str(min_val),
      offset=token_offset,
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=genCtrlDocumentation(df_info_type, str(min_val), allDfInfo, known_df_name, known_col_name, str(min_val), None)
    ),
    CompletionItem(
      value=str(max_val),
      offset=token_offset,
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=genCtrlDocumentation(df_info_type, str(max_val), allDfInfo, known_df_name, known_col_name, str(max_val), None)
    ),
    CompletionItem(
      value=str(mean_val),
      offset=token_offset,
      type=getCompletionItemTypeFromAstPos(need_obj),
      explanation=genCtrlDocumentation(df_info_type, str(mean_val), allDfInfo, known_df_name, known_col_name, str(mean_val), None)
    )
  ]

@typechecked
def cellValueComplete(need_obj: int, token: JupyterlabToken, previousCode: str, analyze_resp: PartialCodeInfo, allDfInfo) -> List[CompletionItem]:
  df_info_type = analyze_resp["df_info_type"]

  if analyze_resp["method_name"] == SptMethodName.VALUE_FILTER:
    df_name = analyze_resp["df_name"]
    col_name = None if analyze_resp["col_name"] is None else removeQuotes(analyze_resp["col_name"])
    cell_value = analyze_resp["cell_value"]
    if (df_name is None) or (col_name is None):
      # debugger.warning("[trivialComplete] cell_value for value_filter should have df_name and col_name")
      return []
    df_info = allDfInfo.get(df_name, {"columns": [], "num_rows": 0, "num_cols": 0})
    col_info = getColInfoByName(df_info["columns"], col_name)
    if col_info is None:
      # debugger.warning(f"[trivialComplete] cell_value for value_filter should have valid col_info, but got {col_info}")
      return []
    if col_info["statsType"] == "categorical":
      if cell_value is None:
        topK = col_info["topK"]
        if topK is None:
          # debugger.warning(f"[trivialComplete] cell_value for value_filter should have valid topK, but got {topK}")
          return []
        return list(map(lambda x: CompletionItem(
          value="\"" + x["value"] + "\"",
          offset=token["offset"],
          type=getCompletionItemTypeFromAstPos(need_obj),
          explanation=genCtrlDocumentation(df_info_type, x["value"], allDfInfo, df_name, col_name, x["value"], None)
        ), topK))
      else:
        return [CompletionItem(
          value=cell_value,
          offset=token["offset"],
          type=getCompletionItemTypeFromAstPos(need_obj),
          explanation=genCtrlDocumentation(df_info_type, cell_value, allDfInfo, df_name, col_name, removeQuotes(cell_value), None)
        )]
    elif col_info["statsType"] == "numeric":
      if cell_value is None:
        return defaultNumCompleteByStats(need_obj, token["offset"], col_info["numStats"], df_info_type, df_name, col_name, allDfInfo)
      else:
        return [CompletionItem(
          value=cell_value,
          offset=token["offset"],
          type=getCompletionItemTypeFromAstPos(need_obj),
          explanation=genCtrlDocumentation(df_info_type, cell_value, allDfInfo, df_name, col_name, cell_value, None)
        )]

  return []

@typechecked
def colIdxComplete(need_obj: int, token: JupyterlabToken, previousCode: str, analyze_resp: PartialCodeInfo, allDfInfo: dtX.TableLevelInfoT) -> List[CompletionItem]:
  lastLineCode = previousCode.split("\n")[-1]
  token_list: List[CompletionItem] = []
  for coln in analyze_resp["col_name_list"]:
    txt = parseGPTX.keepSuffix(lastLineCode, coln)
    realTxt = ""
    if len(lastLineCode) > 0 and lastLineCode[-1] == "'":
      realTxt = txt + "'"
    elif len(lastLineCode) > 0 and lastLineCode[-1] == '"':
      realTxt = txt + '"'
    else:
      realTxt = '"' + txt + '"'


    token_list.append({
      "value": realTxt,
      "offset": token["offset"],
      "type": getCompletionItemTypeFromAstPos(analyze_resp["need_obj"]),
      "explanation": None
    })
  return token_list

@typechecked
def trivialComplete(need_obj: int, token: JupyterlabToken, previousCode: str, analyze_resp: PartialCodeInfo, allDfInfo: dtX.TableLevelInfoT) -> List[CompletionItem]:
  if need_obj == AST_POS.DF_VAR:
    return dfVarComplete(need_obj, token, previousCode, analyze_resp, allDfInfo)
  elif need_obj == AST_POS.PARAM:
    return paramComplete(need_obj, token, previousCode, analyze_resp, allDfInfo)
  elif need_obj == AST_POS.CELL_VALUE:
    return cellValueComplete(need_obj, token, previousCode, analyze_resp, allDfInfo)
  elif need_obj == AST_POS.COL_IDX:
    return colIdxComplete(need_obj, token, previousCode, analyze_resp, allDfInfo)
  return []


import json
from typeguard import typechecked

from .constant import DF_INFO_TYPE
from .datatypes import List, Tuple, Dict, Optional, Union, AllDataFrameStyleT, DocumentationT, ValueShowT, ColumnStyle
from .debugger import debugger
from .utils import fullWordMatch

@typechecked
def initAllDfStyle(allDfInfo, allDfStyle: AllDataFrameStyleT) -> None:
  for dfName in allDfInfo:
    allDfStyle[dfName] = {
      "isFold": True,
      "columns": [],
      "isHidden": False,
    }
    for colobj in allDfInfo[dfName]["columns"]:
      allDfStyle[dfName]["columns"].append({
        "colName": colobj["colName"],
        "isFold": True,
        "isHidden": False,
      })

@typechecked
def multiTableStyleCtrl(co: str, allDfStyle: AllDataFrameStyleT, known_dfName: Optional[str]) -> None:
  need_dfName: List[str] = []
  # 1. match code obj
  for dfName in allDfStyle:
    if fullWordMatch(dfName, co):
      need_dfName.append(dfName)
  # 2. match nearby context
  if known_dfName:
    if known_dfName not in allDfStyle:
      # debugger.warning("[multiTableStyleCtrl] known_dfName not in allDfStyle")
      return
    need_dfName.append(known_dfName)
  # 3. further decide
  if len(need_dfName) == 0:
    # debugger.warning(f"[multiTableStyleCtrl] no dfName found for co: {co}, known_dfName: {known_dfName}.")
    pass
  else:
    for dfName in allDfStyle:
      if dfName in need_dfName:
        continue
      allDfStyle[dfName]["isHidden"] = True

@typechecked
def singleTableStyleCtrl(co: str, allDfStyle: AllDataFrameStyleT, known_dfName: Optional[str]) -> None:
  need_dfName: str = ""
  if known_dfName:
    # 1. match nearby context
    if known_dfName not in allDfStyle:
      # debugger.warning("[singleTableStyleCtrl] known_dfName not in allDfStyle")
      return
    need_dfName = known_dfName
  else:
    # 2. match code obj
    for dfName in allDfStyle:
      if fullWordMatch(dfName, co):
        need_dfName = dfName
        break
  # 3. further decide
  if need_dfName == "":
    # debugger.warning("[singleTableStyleCtrl] no dfName found for co: {co}, known_dfName: {known_dfName}.")
    return
  allDfStyle[need_dfName]["isHidden"] = False
  allDfStyle[need_dfName]["isFold"] = False
  for dfName in allDfStyle:
    if dfName != need_dfName:
      allDfStyle[dfName]["isHidden"] = True
    

@typechecked
def singleColumnStyleCtrl(co: str, allDfStyle: AllDataFrameStyleT, known_dfName: Optional[str], known_colName: Optional[str], known_colIdxList: Optional[str]) -> None:
  need_cols: Dict[str, List[ColumnStyle]] = {}
  # 1. match nearby context
  if known_dfName:
    need_cols[known_dfName] = []
    if known_colName:
      for colobj in allDfStyle[known_dfName]["columns"]:
        if colobj["colName"] == known_colName:
          need_cols[known_dfName].append(colobj)
          break
    if known_colIdxList:
      for colobj in allDfStyle[known_dfName]["columns"]:
        if fullWordMatch(colobj["colName"], known_colIdxList):
          need_cols[known_dfName].append(colobj)
  # 2. match code obj
  for dfName in allDfStyle:
    if fullWordMatch(dfName, co):
      need_cols[dfName] = []
      for colobj in allDfStyle[dfName]["columns"]:
        if fullWordMatch(colobj["colName"], co):
          need_cols[dfName].append(colobj)
  # 3. further decide: no df, or multiple df
  names = list(need_cols.keys())
  if len(names) == 0:
    # debugger.warning(f"[singleColumnStyleCtrl] no dataframe found for co: {co}, known_dfName: {known_dfName}, known_colName: {known_colName}, known_colIdxList: {known_colIdxList}.")
    return
  elif len(names) > 1:
    for dfName in allDfStyle:
      if dfName not in names:
        allDfStyle[dfName]["isHidden"] = True
    return
  # 4. further decide: single df
  for dfName in allDfStyle:
      if dfName not in names:
        allDfStyle[dfName]["isHidden"] = True
      else:
        allDfStyle[dfName]["isHidden"] = False
        allDfStyle[dfName]["isFold"] = False

  colNames = list(map(lambda x: x["colName"], need_cols[names[0]]))
  if len(need_cols[names[0]]) == 0:
    # debugger.warning(f"[singleColumnStyleCtrl] no column found for co: {co}, known_dfName: {known_dfName}, known_colName: {known_colName}, known_colIdxList: {known_colIdxList}.")
    return
  elif len(need_cols[names[0]]) > 1:
    for colobj in allDfStyle[names[0]]["columns"]:
      if colobj["colName"] not in colNames:
        colobj["isHidden"] = True
    return
  else:
    for colobj in allDfStyle[names[0]]["columns"]:
      if colobj["colName"] not in colNames:
        colobj["isHidden"] = True
      else:
        colobj["isHidden"] = False
        colobj["isFold"] = False


@typechecked
def genCtrlDocumentation(df_info_type: str, co: str, allDfInfo, known_dfName: Optional[str], known_colName: Optional[str], known_cellValue: Optional[str], known_colIdxList: Optional[str]) -> str:
  # TODO: add need_obj constraint
  allDfStyle: AllDataFrameStyleT = {}
  initAllDfStyle(allDfInfo, allDfStyle)

  if df_info_type == DF_INFO_TYPE.MULTI_TABLE:
    multiTableStyleCtrl(co, allDfStyle, known_dfName)
  elif df_info_type == DF_INFO_TYPE.SIN_TABLE:
    singleTableStyleCtrl(co, allDfStyle, known_dfName)
  elif df_info_type == DF_INFO_TYPE.MULTI_COL:
    # debugger.warning("[genCtrlDocumentation] DF_INFO_TYPE.MULTI_COL not supported yet")
    pass
  elif df_info_type == DF_INFO_TYPE.SIN_COL:
    singleColumnStyleCtrl(co, allDfStyle, known_dfName, known_colName, known_colIdxList)
  else:
    # debugger.warning(f"[genCtrlDocumentation] not supported df_info_type: {df_info_type}")
    pass
  
  value_show: Union[None, ValueShowT] = None
  if known_dfName and known_colName and known_cellValue:
    value_show = {
      "dfName": known_dfName,
      "colName": known_colName,
      "cellValue": known_cellValue
    }

  doc: DocumentationT = {
    "type": "highlight",
    "explanation": None,
    "highlight": allDfStyle,
    "value_show": value_show,
  }
  doc_str = json.dumps(doc)
  return doc_str

@typechecked
def genExpDocumentation(exp: str) -> str:
  doc: DocumentationT = {
    "type": "explanation",
    "explanation": exp,
    "highlight": None,
    "value_show": None,
  }
  doc_str = json.dumps(doc)
  return doc_str


from typeguard import typechecked
from typing import Union, List, Tuple, Optional

from . import dataframeInfo as dfInfoX
from . import share as shareX
from .. import datatypes as dtX
from .. import utils as utilsX

from ..prompt.fmtCtrl import fmtCtrl_dfVar, fmtCtrl_listOfDf, fmtCtrl_optionalParams, fmtCtrl_selectColumnNames, fmtCtrl_methodRecommendation, fmtCtrl_colVar, fmtCtrl_params, fmtCtrl_binopRHS, fmtCtrl_colIdx, fmtCtrl_aggMethod, fmtCtrl_comment, fmtCtrl_codeLine, fmtCtrl_Param, fmtCtrl_condition
from ..prompt.parseGPTOutput import parseReCodeObj
from ..server.myAIClient import myAIClient
from ..datatypes import PartialCodeInfo, CompletionItem, JupyterlabToken, TableLevelInfoT
from ..constant import AST_POS, PROMPT_SPLITTER, COMPLETE_WHAT, SptMethodName
from ..debugger import debugger

@typechecked
def getFmtCtrlByAstPos(astPosEnum: int, method_name: str):
  completeWhatStr = COMPLETE_WHAT.get(astPosEnum, "")
  if not completeWhatStr:
    # debugger.warning(f"[getFmtCtrlByAstPos] not supported astPosEnum: {astPosEnum}")
    return ""

  if astPosEnum == AST_POS.DF_VAR:
    return fmtCtrl_dfVar(completeWhatStr)
  elif astPosEnum == AST_POS.LIST_DF_VARS:
    return fmtCtrl_listOfDf(completeWhatStr)
  elif astPosEnum == AST_POS.LIST_COL_IDX:
    return fmtCtrl_selectColumnNames(completeWhatStr)
  elif astPosEnum == AST_POS.OPT_PARAM:
    return fmtCtrl_optionalParams(method_name, completeWhatStr)
  elif astPosEnum == AST_POS.PD_FUNC_NAME or astPosEnum == AST_POS.DF_METHOD_NAME or astPosEnum == AST_POS.COL_METHOD_NAME:
    return fmtCtrl_methodRecommendation(completeWhatStr)
  elif astPosEnum == AST_POS.COL_VAR:
    return fmtCtrl_colVar(completeWhatStr)
  elif astPosEnum == AST_POS.PARAM:
    return fmtCtrl_Param()
    # return fmtCtrl_params(completeWhatStr)
  elif astPosEnum == AST_POS.BINOP_RHS:
    return fmtCtrl_binopRHS(completeWhatStr)
  elif astPosEnum == AST_POS.COL_IDX:
    return fmtCtrl_colIdx(completeWhatStr)
  elif astPosEnum == AST_POS.AGG_METHOD_NAME:
    return fmtCtrl_aggMethod(completeWhatStr)
  elif astPosEnum == AST_POS.COMMENT:
    return fmtCtrl_comment(completeWhatStr)
  elif astPosEnum == AST_POS.CODE_LINE:
    return fmtCtrl_codeLine(completeWhatStr)
  elif astPosEnum == AST_POS.CONDITION:
    return fmtCtrl_condition(completeWhatStr)
  else:
    # debugger.warning(f"[getFmtCtrlByAstPos] not supported astPosEnum: {astPosEnum}")
    return ""

# SC = special case
@typechecked
def gptCompleteForSC(pCode: str, method_name: str = "", need_obj: int = -1, tableLvInfo: Optional[TableLevelInfoT] = None, columnLvInfo = None, rowLvInfo = None) -> Optional[str]:
  # get codeInfo
  codeInfo = shareX.getCodeInfo(pCode)
  # get dfInfo
  dfInfoPrompt: str = f"{dfInfoX.dataInfoPreamble()}\n"
  if (tableLvInfo is None) and (columnLvInfo is None) and (rowLvInfo is None):
    # debugger.warning(f"[gptCompleteForSC] data context not specified")
    return None
  if tableLvInfo is not None:
    dfInfoPrompt += dfInfoX.multiTableLvInfoPrompt(tableLvInfo)
  if columnLvInfo is not None:
    dfInfoPrompt += f"""\n{dfInfoX.multiColLvInfoPrompt(columnLvInfo, tableLvInfo)}"""
  if rowLvInfo is not None:
    dfInfoPrompt += f"""\n{dfInfoX.multiRowLvInfoPrompt(rowLvInfo)}"""
  
  # generate prompt
  prompt = codeInfo + PROMPT_SPLITTER
  prompt += dfInfoPrompt + PROMPT_SPLITTER
  prompt += getFmtCtrlByAstPos(need_obj, method_name)

  return prompt


@typechecked
def gptComplete_Exe(client: myAIClient, token: JupyterlabToken, prompt: str, pCode: str, need_obj: int) -> List[CompletionItem]:
  token_list: List[CompletionItem] = []
  resp = client.sendPrompt(prompt)
  lastLineCode = pCode.split("\n")[-1]
  pres = parseReCodeObj(resp, lastLineCode)
  for i, co in enumerate(pres):
    token_list.append({
      "value": co,
      "offset": token["offset"],
      "type": utilsX.getCompletionItemTypeFromAstPos(AST_POS.CODE_LINE),
      "explanation": None
    })
  return token_list

@typechecked
def filterDataContext(arg: PartialCodeInfo, tableLvInfo: Optional[dtX.TableLevelInfoT], colLvInfo: Optional[dtX.ColLevelInfoT] = None, rowLvInfo: Optional[dtX.RowLevelInfoT] = None) -> Tuple[Optional[dtX.TableLevelInfoT], Optional[dtX.ColLevelInfoT], Optional[dtX.RowLevelInfoT]]:
  method_name = arg["method_name"]
  need_obj = arg["need_obj"]
  df_name = arg["df_name"]
  col_name = arg["col_name"]
  filteredTLI: Optional[dtX.TableLevelInfoT] = None
  filteredCLI: Optional[dtX.ColLevelInfoT] = None
  filteredRLI: Optional[dtX.RowLevelInfoT] = None
  if (not method_name) or (need_obj < 0):
    # debugger.warning(f"[filterDataContext] method_name {method_name} or need_obj {need_obj} not specified")
    return filteredTLI, filteredCLI, filteredRLI
  
  if method_name == SptMethodName.DF_COL or method_name == SptMethodName.COL_METHOD_NAME:
    filteredTLI = tableLvInfo
    if colLvInfo is not None and colLvInfo.get(df_name) is not None and colLvInfo[df_name].get(col_name) is not None:
      infoItem: dtX._ColLevelInfoT = colLvInfo[df_name][col_name]
      filteredCLI = {
        df_name: {
          col_name: {
            "dtype": infoItem["dtype"],
            "nullCount": infoItem["nullCount"],
            "sortedness": infoItem["sortedness"],
            "cardinality": infoItem["cardinality"],
            "uniqueValues": infoItem["uniqueValues"][:50] if infoItem["uniqueValues"] is not None else None,
            "uvCounts": None,
            "minValue": infoItem["minValue"],
            "maxValue": infoItem["maxValue"],
            "sample": infoItem["sample"][:50] if infoItem["sample"] is not None else None
          }
        }
      }
    filteredRLI = None
  elif method_name in [SptMethodName.COL_STR_REPLACE, SptMethodName.COL_FILLNA, SptMethodName.COL_REPLACE, SptMethodName.VALUE_FILTER]:
    filteredTLI = None
    if colLvInfo is not None and colLvInfo.get(df_name) is not None and colLvInfo[df_name].get(col_name) is not None:
      infoItem: dtX._ColLevelInfoT = colLvInfo[df_name][col_name]
      filteredCLI = {
        df_name: {
          col_name: {
            "dtype": infoItem["dtype"],
            "nullCount": infoItem["nullCount"],
            "sortedness": None,
            "cardinality": None,
            "uniqueValues": infoItem["uniqueValues"][:50] if infoItem["uniqueValues"] is not None else None,
            "uvCounts": None,
            "minValue": infoItem["minValue"],
            "maxValue": infoItem["maxValue"],
            "sample": infoItem["sample"][:50] if infoItem["sample"] is not None else None
          }
        }
      }
    filteredRLI = None
  elif method_name in [SptMethodName.DF_SORT_VALUES, SptMethodName.DF_DROP_DUPLICATES, SptMethodName.DF_DROP, SptMethodName.DF_RENAME, SptMethodName.DF_MELT, SptMethodName.DF_PIVOT, SptMethodName.DF_PIVOT_TABLE]:
    filteredTLI = None
    filteredCLI = None
    if rowLvInfo is not None and rowLvInfo.get(df_name) is not None:
      filteredRLI = {
        df_name: rowLvInfo[df_name]
      }
  elif method_name in [SptMethodName.DF_DROPNA]:
    if tableLvInfo is not None and tableLvInfo.get(df_name) is not None:
      filteredTLI = {
        df_name: {
          "numRows": None,
          "numCols": None,
          "columnNameList": tableLvInfo[df_name]["columnNameList"]
        }
      }
    if colLvInfo is not None and colLvInfo.get(df_name) is not None:
      filteredCLI = {
        df_name: {}
      }
      for col in colLvInfo[df_name]:
        filteredCLI[df_name][col] = {
          "dtype": colLvInfo[df_name][col]["dtype"],
          "nullCount": colLvInfo[df_name][col]["nullCount"],
          "sortedness": None,
          "cardinality": None,
          "uniqueValues": None,
          "uvCounts": None,
          "minValue": None,
          "maxValue": None,
          "sample": None
        }
  elif method_name == SptMethodName.DF_GROUPBY:
    if tableLvInfo is not None and tableLvInfo.get(df_name) is not None:
      filteredTLI = {
        df_name: tableLvInfo[df_name]
      }
    filteredCLI = {}
    if colLvInfo is not None and colLvInfo.get(df_name) is not None:
      filteredCLI[df_name] = {}
      dfObj = colLvInfo[df_name]
      for col in dfObj:
        # TODO: We also include unknown type
        if dfObj[col].get("dtype") is None or utilsX.isCateColumn(dfObj[col]["dtype"]):
          filteredCLI[df_name][col] = {
            "dtype": dfObj[col]["dtype"],
            "nullCount": None,
            "sortedness": None,
            "cardinality": dfObj[col]["cardinality"],
            "uniqueValues": None,
            "uvCounts": None,
            "minValue": None,
            "maxValue": None,
            "sample": None
          }
    if rowLvInfo is not None and rowLvInfo.get(df_name) is not None:
      filteredRLI = {
        df_name: {
          "columnNameList": rowLvInfo[df_name]["columnNameList"],
          "sampleRows": rowLvInfo[df_name]["sampleRows"][:5] if rowLvInfo[df_name]["sampleRows"] is not None else None
        }
      }
  elif method_name == SptMethodName.COLUMNS_SELECT:
    if tableLvInfo is not None and tableLvInfo.get(df_name) is not None:
      filteredTLI = {
        df_name: {
          "numRows": None,
          "numCols": None,
          "columnNameList": tableLvInfo[df_name]["columnNameList"]
        }
      }
    filteredCLI = None
    filteredRLI = None
  elif method_name == SptMethodName.ASSIGN_STMT:
    filteredTLI = tableLvInfo
    if col_name is not None and colLvInfo is not None and colLvInfo.get(df_name) is not None and colLvInfo[df_name].get(col_name) is not None:
      filteredCLI = {
        df_name: {
          col_name: {
            "dtype": colLvInfo[df_name][col_name]["dtype"],
            "nullCount": colLvInfo[df_name][col_name]["nullCount"],
            "sortedness": colLvInfo[df_name][col_name]["sortedness"],
            "cardinality": colLvInfo[df_name][col_name]["cardinality"],
            "uniqueValues": colLvInfo[df_name][col_name]["uniqueValues"][:50] if colLvInfo[df_name][col_name]["uniqueValues"] is not None else None,
            "uvCounts": None,
            "minValue": colLvInfo[df_name][col_name]["minValue"],
            "maxValue": colLvInfo[df_name][col_name]["maxValue"],
            "sample": colLvInfo[df_name][col_name]["sample"][:50] if colLvInfo[df_name][col_name]["sample"] is not None else None
          }
        }
      }
    if filteredCLI is None:
      filteredRLI = {}
      for dfName in rowLvInfo:
        filteredRLI[dfName] = {
          "columnNameList": rowLvInfo[dfName]["columnNameList"],
          "sampleRows": rowLvInfo[dfName]["sampleRows"][:5] if rowLvInfo[dfName]["sampleRows"] is not None else None
        }
  elif method_name == SptMethodName.COMMENT or method_name == SptMethodName.CODE_LINE or method_name == SptMethodName.DF_MERGE or method_name == SptMethodName.DF_JOIN or method_name == SptMethodName.PD_MERGE:
    filteredTLI = tableLvInfo
    filteredCLI = None
    filteredRLI = {}
    for dfName in rowLvInfo:
      filteredRLI[dfName] = {
        "columnNameList": rowLvInfo[dfName]["columnNameList"],
        "sampleRows": rowLvInfo[dfName]["sampleRows"][:5] if rowLvInfo[dfName]["sampleRows"] is not None else None
      }


  return filteredTLI, filteredCLI, filteredRLI




from typing import List, Tuple, Dict, Set, Union, Optional, Literal, TypedDict, Any

class TokenInfo(TypedDict):
    type: int
    exact_type: int
    string: str

# Note:
# 1. `df_name` is only available when the cursor is next to a dataframe name. e.g. `df.`
# 2. `method_name` also includes things like "column selection"    

class PartialCodeInfo(TypedDict):
    method_name: Optional[str]
    need_obj: int
    df_info_type: str
    df_name: Optional[str]
    df_name_list: Optional[List[str]]
    col_name: Optional[str]
    col_name_list: Optional[List[str]]
    col_name_quote: Optional[str]
    op_name: Optional[str]
    cell_value: Optional[str]
    col_idx_list: Optional[str]
    special_case: int
    is_trivial: bool

class CompletionItem(TypedDict):
    value: str
    offset: int
    type: str
    explanation: Optional[str]

# Middle layer
class _dfInfo_listDf(TypedDict):
    shape: str
    columns: str

# Suitable for list of dataframes
dfInfo_listDf = Dict[str, _dfInfo_listDf]

# Middle layer
# Cannot written as TypedDict because of "25%", "50%", "75%" is not a valid variable name


SortType = Literal['ascending', 'descending', 'noSort']

class NumericalStatsT(TypedDict):
    min: Union[float, int]
    q25: Union[float, int]
    q50: Union[float, int]
    q75: Union[float, int]
    max: Union[float, int]
    mean: Union[float, int]
    count: int
    std: Union[float, int]
    sd_outlier: int
    iqr_outlier: int
    sortedness: SortType
    n_zero: int
    n_positive: int
    n_negative: int

class BinInfo(TypedDict):
    left: Union[int, float]
    right: Union[int, float]
    closed_left: bool
    closed_right: bool
    count: int

class CategoricalStatsT(TypedDict):
    minLength: int
    maxLength: int
    meanLength: Union[float, int]
    cardinality: int

class TemporalStatsT(TypedDict):
    num_outliers: int
    sortedness: SortType

class TopKItemT(TypedDict):
    value: str
    count: int

class TempIntervalT(TypedDict):
    months: int
    days: int
    micros: int

class TempHistItemT(TypedDict):
    bucket: int
    low: Union[int, float]
    high: Union[int, float]
    count: int

# Middle layer
class ColumnInfo(TypedDict):
    colName: str
    dtype: str
    nullCount: int
    statsType: Literal["numeric", "categorical", "temporal", "boolean", "other"]
    # Pitfall: Union[None, T] != Optional[T], especially when T is a TypedDict
    numStats: Union[None, NumericalStatsT]
    cateStats: Union[None, CategoricalStatsT]
    timeStats: Union[None, TemporalStatsT]
    bins: Optional[List[BinInfo]]
    topK: Optional[List[TopKItemT]]
    tempHist: Optional[List[TempHistItemT]]
    tempInterval: Union[None, TempIntervalT]

class JupyterlabToken(TypedDict):
    value: str
    offset: int
    type: Optional[str]

class MultiLevelPrefix(TypedDict):
    start_symbol: str
    delimiter: str
    index_range: List[int]

class SpecialTokenMapping(TypedDict):
    key: str
    value: Union[str, List[str]]

class SigInfo(TypedDict):
    module_name: str
    name: str
    full_name: str
    type: str

class ColumnStyle(TypedDict):
    colName: str
    isFold: bool
    isHidden: bool

class _DataFrameStyle(TypedDict):
    isFold: bool
    isHidden: bool
    columns: List[ColumnStyle]

AllDataFrameStyleT = Dict[str, _DataFrameStyle]

class ValueShowT(TypedDict):
    dfName: str
    colName: str
    cellValue: str

class DocumentationT(TypedDict):
    type: Literal["explanation", "highlight"]
    explanation: Optional[str]
    highlight: Union[None, AllDataFrameStyleT]
    value_show: Union[None, ValueShowT]

class _TableLevelInfoT(TypedDict):
    columnNameList: List[str]
    numRows: Optional[int]
    numCols: Optional[int]

TSortedness = Literal["asc", "desc", "no"]

class _ColLevelInfoT(TypedDict):
    # Common
    dtype: Optional[str]
    nullCount: Optional[int]
    sortedness: Optional[TSortedness]
    # For categorical columns
    cardinality: Optional[int] # Number of unique values
    uniqueValues: Optional[List[Union[str, int, float]]]; # At most 1000 unique values
    uvCounts: Optional[List[int]]; # If not null, must have the same length as uniqueValues
    # For numerical columns
    minValue: Union[None, int, float, str]
    maxValue: Union[None, int, float, str]
    sample: Optional[List[Union[int, float, str]]]

DFRowT = Dict[str, Union[int, float, bool, str, None]]

class _RowLevelInfoT(TypedDict):
    columnNameList: List[str]
    sampleRows: Optional[List[DFRowT]]

TableLevelInfoT = Dict[str, _TableLevelInfoT]
RowLevelInfoT = Dict[str, _RowLevelInfoT]
ColLevelInfoT = Dict[str, Dict[str, _ColLevelInfoT]]


from typeguard import typechecked
from ..constant import PROMPT_MARKER

@typechecked
def fmtCtrl_basedOnWhat(provideExample: bool = True) -> str:
  example = "(E.g. schema of the table, shape of the DataFrame, data type of the columns, distribution of each column, data quality issues, special data values in the column, etc.)"
  return f"""So, based on the {PROMPT_MARKER.CODE_CONTEXT} and {PROMPT_MARKER.DATA_CONTEXT}{" " + example if provideExample else ""} provided above"""

@typechecked
def fmtCtrl_forExample() -> str:
  return f"""For example, if the user is selecting a subset of the dataframe or columns using symbols like `[`, you may need to recommend methods that returns a boolean mask in order to filter the data. Meanwhile, such a filtering may originate from some characteristics of the data, such as statistical distribution or quality issues."""

@typechecked
def fmtCtrl_noExplanation(noNeedWholeLine: bool = True) -> str:
  noNeedWholeLineStr = "(it is not necessary to complete the whole line of code)"
  return f"""without explanation{f" {noNeedWholeLineStr}" if noNeedWholeLine else ""}"""

@typechecked
def fmtCtrl_additionalNotes(wrapWhat: str, multipleItems: bool = True, forbiddenFuncStr: str = "") -> str:
  optionalStr = "You don't need to put too many items in an option and don't make an option too long. "
  resStr = f"""You can list more than one option (5 options at most) if you think there are many possibilities, and order them, putting the most possible one to the top.{" " + forbiddenFuncStr if forbiddenFuncStr != "" else ""} You don't need to use backtick to wrap the {wrapWhat}."""
  if multipleItems:
    resStr = optionalStr + resStr
  return resStr

@typechecked
def fmtCtrl_specialCase() -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please provide the autocompletion prompt for the user, which consists of the expression and explanation of your choice (No need to make the explanation too long and it would be better if the explanation is related to the data context. For instance you can mention the name of the dataframe variable or the name of the column). {fmtCtrl_noExplanation()}. The output format is like:
1. expression1 | explanation1
2. expression2 | explanation2
...
{fmtCtrl_additionalNotes("expression")}"""

@typechecked
def fmtCtrl_dfVar(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the dataframes appearing in this code script. (2) order these dataframes based on some criteria like similar schema, similar semantics or other common features with the dataframe nearby the lastline of code, (3) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. df1
2. df2
...
{fmtCtrl_additionalNotes(completeWhat, False)}"""

@typechecked
def fmtCtrl_listOfDf(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the dataframes appearing in this code script, (2) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. [df1_1, df1_2, ...]
2. [df2_1, df2_2, ...]
...
{fmtCtrl_additionalNotes(completeWhat)}"""

@typechecked
def fmtCtrl_selectColumnNames(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) think about the intents of the data worker from the previous code (That is because data workers usually choose columns from a large table for a more concise view and better analysis between several columns). (2) choose some combinations of several column names appearing in this dataframe, (3) provide the autocompletion prompt for the user (don't select too much columns. Sometimes one column is also OK), {fmtCtrl_noExplanation()}. The output format is like:
1. "column1_1", "column1_2", ...
2. "column2_1", ...
...
{fmtCtrl_additionalNotes(completeWhat)}"""

@typechecked
def fmtCtrl_optionalParams(method_name: str, completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the optional parameters available and not specified yet in the current signature `{method_name}`, (2) provide approriate value for each parameter you select, (3) output the corresponding autocompletion prompts, {fmtCtrl_noExplanation()}. The output format is like:
1. parameter1_1=value1_1, parameter1_2=value1_2, ...
2. parameter2_1=value2_1, ...
...
{fmtCtrl_additionalNotes(completeWhat)}"""

@typechecked
def fmtCtrl_methodRecommendation(completeWhat: str) -> str:
  """
  Previously, the prompt is:
{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the methods available in these categories listed above (you need to carefully read the explanation of each category), (2) order theses methods based on the correlation among the description of the corresponding category, the current code and data context ({fmtCtrl_forExample()}), (3) provide the autocompletion prompt for the user, which consists of the method name and explanation of your choice (No need to make the explanation too long and it would be better if the explanation is related to the data context. For instance you can mention the name of the dataframe variable or the name of the column). The output format is like:
1. method_name_1 | explanation_1
2. method_name_2 | explanation_2
...
{fmtCtrl_additionalNotes(completeWhat, False)}
"""

  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the methods available in these categories listed above (you need to carefully read the explanation of each category), (2) order theses methods based on the correlation among the description of the corresponding category, the current code and data context ({fmtCtrl_forExample()}), (3) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. method_name_1
2. method_name_2
...
{fmtCtrl_additionalNotes(completeWhat, False)}"""

@typechecked
def fmtCtrl_colVar(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the column names appearing in this dataframe or other dataframes. Note that you may first consider the columns with the more detailed information, and then move to the less detailed ones if such a column is not appropriate, (2) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. df1["column1_1"]
2. df1["column1_2"]
3. df2["column2_1"]
...
{fmtCtrl_additionalNotes(completeWhat, False)}"""

@typechecked
def fmtCtrl_params(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) think of the possible combinations of these mandatory parameters, (2) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. parameter1_1=value1_1, parameter1_2=value1_2, ...
2. parameter2_1=value2_1, ...
...
{fmtCtrl_additionalNotes(completeWhat)}"""

@typechecked
def fmtCtrl_binopRHS(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) Carefully read the documentation mentioned above, (2) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. <OP1> <RHS1>
2. <OP2> <RHS2>
...
Where OP can be "+", "-", "*", "/", ">", "<", ">=", "<=", "==", "!=", and the RHS can be a scalar, a pandas Series. {fmtCtrl_additionalNotes(completeWhat, False)}"""

@typechecked
def fmtCtrl_colIdx(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a column name based on the context near the last line of code (i.e. you need to make sure that the column can be accessed in the current dataframe), (2) order these column names you select based on several criteria (for example, if the column has been mentioned above, it is likely to be accessed in the following code), (3) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. "column_name_1"
2. "column_name_2"
...
{fmtCtrl_additionalNotes(completeWhat, False)}"""

@typechecked
def fmtCtrl_aggMethod(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) select a subset of the aggregation methods available in documentation, (2) order these methods based on several criteria (for example, Sometimes you may also need to notice the the name of the variable before "=" in the last line of the incomplete code, because the meaning of variable name may be a useful clue to guess the appropriate aggregation method name) (3) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. method_name1()
2. method_name2()
...
{fmtCtrl_additionalNotes(completeWhat, False)}"""

@typechecked
def fmtCtrl_comment(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat()}, please: (1) think of the possible comments that can be added to the code, (2) order these comments based on the correlation between the comments and "{PROMPT_MARKER.CODE_CONTEXT} and {PROMPT_MARKER.DATA_CONTEXT}", (3) provide the autocompletion prompt for the user, {fmtCtrl_noExplanation()}. The output format is like:
1. <comment 1>
2. <comment 2>
...
{fmtCtrl_additionalNotes(completeWhat)}"""

@typechecked
def forbiddenFunc() -> str:
  return f"""Please note that you should not provide the following autocompletion prompts: .head(), .info(), .describe(), .shape, .plot(), .value_counts(), .unique(), nunique(), .nlargest()."""

@typechecked
def fmtCtrl_codeLine(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat(False)}, please: (1) think of the possible codes (e.g. data transformation operations applied to the datasets) the user wants to write; (2) rank these codes based on the correlation between the code and "{PROMPT_MARKER.CODE_CONTEXT} and {PROMPT_MARKER.DATA_CONTEXT}"; (3) provide the autocompletion prompt and add to {PROMPT_MARKER.CURSOR} of the code, {fmtCtrl_noExplanation(False)}. The output format is like:
1. <rest part of code line 1>
2. <rest part of code line 2>
...
{fmtCtrl_additionalNotes(completeWhat, False, forbiddenFunc())}"""

@typechecked
def fmtCtrl_condition(completeWhat: str) -> str:
  return f"""{fmtCtrl_basedOnWhat(False)}, please: (1) think of the possible operators and operands that can be applied to filter the data, (2) rank these operators and operands based on the correlation between the condition and "{PROMPT_MARKER.CODE_CONTEXT} and {PROMPT_MARKER.DATA_CONTEXT}", (3) provide the autocompletion prompt and add to {PROMPT_MARKER.CURSOR} of the code, {fmtCtrl_noExplanation(False)}. The output format is like:
1. <the rest part of condition 1>
2. <the rest part of condition 2>
...
{fmtCtrl_additionalNotes(completeWhat, False)}""" 

@typechecked
def fmtCtrl_Param() -> str:
  return f"""{fmtCtrl_basedOnWhat(False)}, please: (1) think of the usage of the function and the possible parameters that can be passed to the function, (2) think of the possible parameters values and rank these parameters based on the correlation between the parameter and "{PROMPT_MARKER.CODE_CONTEXT} and {PROMPT_MARKER.DATA_CONTEXT}", (3) provide the autocompletion prompt and add to {PROMPT_MARKER.CURSOR} of the code, {fmtCtrl_noExplanation(False)}. The output format is like:
1. <code line 1>
2. <code line 2>
...
{fmtCtrl_additionalNotes("code line", False)}"""


import json
import traceback
from typeguard import typechecked

from .server.server import apiFunc as apiX
from .server.server.myAIClient import myAIClient
from .server.debugger import debugger
from .server.constant import LLAMA3_70B_8192, LLAMA31_70B_VERSATILE, RUN_HOST, RUN_PORT

CLIENT = myAIClient(LLAMA3_70B_8192)

@typechecked
def setGroqAPIKey(apiKey: str) -> None:
	global CLIENT
	res = {
		"status": True,
	}
	if not apiKey:
		debugger.error("API Key is empty. Please set a valid API Key.")
		res["status"] = False
		print(json.dumps(res, default=str))
	CLIENT.initGroqClient(apiKey)
	print(json.dumps(res, default=str))
	
@typechecked
def complete(_para: str) -> None:
	global CLIENT
	para = json.loads(_para)
	previousCode2D = para["previousCode2D"]
	token = para["token"]
	tableLvInfo = para["tableLvInfo"]
	colLvInfo = para["colLvInfo"]
	rowLvInfo = para["rowLvInfo"]

	resJson = {
		"tokenList": [],
		"analyzeResp": None
	}
	try: 
		token_list, analyze_resp = apiX.try_complete(CLIENT, previousCode2D, token, tableLvInfo, colLvInfo, rowLvInfo)
		resJson = {
			"tokenList": token_list,
			"analyzeResp": analyze_resp
		}
	except Exception as e:
		debugger.error(f"Error in complete: {str(e)}")
		traceback.print_exc()

	print(json.dumps(resJson, default=str))


import os
from typeguard import typechecked
from typing import List, Dict, Optional, Union
from openai import OpenAI

from .. import constant as constX
from ..debugger import debugger
from .. import utils as utilsX

@typechecked
class myAIClient():
  def __init__(self, model: str) -> None:
    self.chatHistory: List[Dict[str, str]] = []
    self.model = model
    self.groqClient: Optional[OpenAI] = None

  def promptRemote(self, prompt: str, client: OpenAI) -> str:
    self.chatHistory = [{
      "role": "user",
      "content": prompt
    }]

    response = None
    text: Optional[str] = None
    try:
      response = client.chat.completions.create(
        messages=self.chatHistory,
        model=self.model,
        max_tokens=constX.MAX_OUTPUT_TOKENS,
        temperature=constX.TEMPERATURE
      )
    except Exception as e:
      debugger.error(f"[promptRemote] {e}\n\n\nOriginal Prompt:\n{prompt}")
      return ""

    text = response.choices[0].message.content
    if text is None:
      raise RuntimeError("[promptRemote] No text provided")

    return text

  def sendPrompt(self, prompt: str) -> str:
    utilsX.logInFiles("---------\nPrompt: " + prompt + "\n---------")
    if not self.groqClient:
      raise RuntimeError("Groq API key is not set. Please set the API key before making requests.")
    
    # OpenAI(
      # base_url="https://api.groq.com/openai/v1",
    # )

    text = self.promptRemote(prompt, self.groqClient)
    
    utilsX.logInFiles("\n--------response--------\n" + text)
    utilsX.logInFiles("--------finished--------")

    return text

  def initGroqClient(self, apiKey: str) -> None:
    self.groqClient = OpenAI(
      base_url="https://api.groq.com/openai/v1",
      api_key=apiKey
    )


import re
import tokenize
from typing import List, Tuple, Optional
from typeguard import typechecked

from .. import datatypes as dtX


@typechecked
def get_signature_fixjedi(lastline: str, pandasAlias: str) -> List[dtX.SigInfo]:
  sigs: List[dtX.SigInfo] = []
  matchres = closestSignature(lastline)
  if matchres is None:
    return sigs
  if matchres[0] in ["groupby", "join", "sort_values", "drop_duplicates", "drop", "rename", "dropna", "pivot", "pivot_table", "melt"]: # "pop", 
    sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.frame", full_name=f"pandas.core.frame.DataFrame.{matchres[0]}"))
  elif matchres[0] in ["contains", "endswith", "startswith", "removeprefix", "removesuffix", "strip", "lstrip", "rstrip", "extract", "extractall", "fullmatch", "find", "rfind", "findall", "slice"]:
    # if there are ".str." before the match
    if lastline[:matchres[1]].endswith("].str."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.strings.accessor", full_name=f"pandas.core.strings.accessor.StringMethods.{matchres[0]}"))
  elif matchres[0] in ["astype", "fillna", "isin"]:
    if lastline[:matchres[1]].endswith("]."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.series", full_name=f"pandas.core.series.Series.{matchres[0]}"))
  elif matchres[0] in ["concat"]:
    if lastline[:matchres[1]].endswith(f"{pandasAlias}."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.reshape", full_name=f"pandas.core.reshape.{matchres[0]}"))
  elif matchres[0] in ["replace"]:
    if lastline[:matchres[1]].endswith("]."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.series", full_name=f"pandas.core.series.Series.{matchres[0]}"))
    elif lastline[:matchres[1]].endswith("].str."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.strings.accessor", full_name=f"pandas.core.strings.accessor.StringMethods.{matchres[0]}"))
  elif matchres[0] in ["merge"]:
    if lastline[:matchres[1]].endswith(f"{pandasAlias}."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.reshape", full_name=f"pandas.core.reshape.{matchres[0]}"))
    elif lastline[:matchres[1]].endswith("."):
      sigs.append(dtX.SigInfo(name=matchres[0], type="function", module_name="pandas.core.frame", full_name=f"pandas.core.frame.DataFrame.{matchres[0]}"))
  return sigs

     

@typechecked
def closestSignature(text_before_cursor: str) -> Optional[Tuple[str, int]]:
  # Regex pattern to match function calls with an open parenthesis but no closing parenthesis
  pattern_open = r'\b(\w+)\s*\([^)]*$'
  
  # Check if the cursor is within an open function call
  if re.search(pattern_open, text_before_cursor):
      # Regex pattern to find all function calls
      pattern_all = r'\b(\w+)\s*\('
      # Find all matches in the text
      matches = [(m.group(1), m.start(1)) for m in re.finditer(pattern_all, text_before_cursor)]
      # Return the last match if any, otherwise return None
      if matches:
          return matches[-1]
  # If no open function call is found, return None
  return None


class AST_POS:
  INVALID = -1
  GLOBAL = 0
  SIGNATURE = 1

  DF_VAR = 10
  LIST_DF_VARS = 11
  LIST_COL_IDX = 12
  OPT_PARAM = 13
  COL_VAR = 14
  PARAM = 15
  BINOP_RHS = 16
  COL_IDX = 17
  CELL_VALUE = 18

  COMMENT = 30
  CODE_LINE = 31
  CONDITION = 32

  PD_FUNC_NAME = 100
  DF_METHOD_NAME = 101
  COL_METHOD_NAME = 102
  AGG_METHOD_NAME = 103

class SPECIAL_CASE:
  NONE = 0
  DF_NAME_PREFIX = 2 # E.g. `covi` of dataframe `covid_19_data`
  SIN_COL_INDEX = 3
  DF_SELECT = 4 # E.g. `covid_19_data[`
  DF_SELECT_COL_1 = 5 # E.g. `covid_19_data["`
  DF_SELECT_COL_2 = 6 # E.g. `covid_19_data["Cou` of `covid_19_data["Country / Region"]`
  DF_SELECT_COL_3 = 7 # E.g. `"Country /`


class DF_INFO_TYPE:
  MULTI_TABLE = "multi_table"
  SIN_TABLE = "single_table"
  MULTI_COL = "multi_column"
  SIN_COL = "single_column"
  USER_SPECIFIED = "user_specified"


class AnalyzeClasses:
  GLOBAL = "global"
  IN_PD_FUNC_SIG = "inPDFuncSig"
  IN_DF_METHOD_SIG = "inDFMethodSig"
  IN_COL_METHOD_SIG = "inColMethodSig"
  IN_COL_STR_METHOD_SIG = "inColStrMethodSig"

class AnalyzeEntry:
  RULES = "rules"
  FULL_NAME = "fullName"
  ML_PREFIX = "multiLevelPrefixes"
  LL_PREFIX = "lastLevelPrefixes"
  NEED_OBJ = "needObj"
  DF_INFO_TYPE = "dfInfoType"
  IS_TRIVIAL = "isTrivial"

class SpecialTokens:
  PANDAS = "$$pandas"
  DF = "$$df"
  COLIDX = "$$colidx"
  COLIDX_LIST = "$$colidx_list"
  OP = "$$op"
  CELL_VALUE = "$$cell_value"

# support method full name
class SptMethodName:
  # special
  PD_FUNC_NAME = "pdFuncName"
  DF_METHOD_NAME = "dfMethodName"
  COL_METHOD_NAME = "colMethodName"
  DF_COL = "dfCol" # E.g. `df["col_name"]`
  COLUMNS_SELECT = "columnsSelect"
  GROUPBY_COLSELECT = "groupbyColSelect"
  GROUPBY_AGGNAME = "groupbyAggName"
  VALUE_FILTER = "valueFilter"
  EXPR = "expr"
  COMMENT = "comment"
  CODE_LINE = "codeLine"
  PREFIX_DF_NAME = "prefixDfName"
  PREFIX_COL_NAME = "prefixColName"
  DF_SELECT = "dfSelect"
  ASSIGN_STMT = "assignStmt"
  # pandas methods
  PD_CONCAT = "pandas.concat"
  PD_MERGE = "pandas.merge"
  # pandas.DataFrame methods
  DF_MERGE = "pandas.DataFrame.merge"
  DF_DROP = "pandas.DataFrame.drop"
  DF_GROUPBY = "pandas.DataFrame.groupby"
  DF_JOIN = "pandas.DataFrame.join"
  DF_SORT_VALUES = "pandas.DataFrame.sort_values"
  DF_DROP_DUPLICATES = "pandas.DataFrame.drop_duplicates"
  DF_MELT = "pandas.DataFrame.melt"
  DF_RENAME = "pandas.DataFrame.rename"
  DF_DROPNA = "pandas.DataFrame.dropna"
  DF_PIVOT = "pandas.DataFrame.pivot"
  DF_PIVOT_TABLE = "pandas.DataFrame.pivot_table"
  # pandas.Series methods
  COL_ISIN = "pandas.Series.isin"
  COL_ASTYPE = "pandas.core.series.Series.astype"
  COL_FILLNA = "pandas.core.series.Series.fillna"
  COL_REPLACE = "pandas.core.series.Series.replace"
  # pandas
  COL_STR_REPLACE = "pandas.core.strings.accessor.StringMethods.replace"


PROMPT_SPLITTER = "\n--------------------\n\n"
LLAMA3_70B_8192 ="llama3-70b-8192"
LLAMA31_70B_VERSATILE = "llama-3.1-70b-versatile"
RUN_HOST = "0.0.0.0"
RUN_PORT = 1022
MAX_OUTPUT_TOKENS = 70
TEMPERATURE = 0.01

COMPLETION_ITEM_KIND = {
  AST_POS.DF_VAR: "df",
  AST_POS.LIST_DF_VARS: "df",
  AST_POS.LIST_COL_IDX: "col",
  AST_POS.OPT_PARAM: "param",
  AST_POS.PD_FUNC_NAME: "func",
  AST_POS.DF_METHOD_NAME: "func",
  AST_POS.COL_METHOD_NAME: "func",
  AST_POS.COL_VAR: "col",
  AST_POS.PARAM: "param",
  AST_POS.BINOP_RHS: "calc",
  AST_POS.COL_IDX: "col",
  AST_POS.CELL_VALUE: "cell",
  AST_POS.AGG_METHOD_NAME: "func",
  AST_POS.CODE_LINE: "code",
}

COMPLETE_WHAT = {
  AST_POS.DF_VAR: "dataframe variable",
  AST_POS.LIST_DF_VARS: "list of dataframes",
  AST_POS.LIST_COL_IDX: "list of column indices",
  AST_POS.OPT_PARAM: "optional parameters",
  AST_POS.PD_FUNC_NAME: "function name of pandas module",
  AST_POS.DF_METHOD_NAME: "method name of the dataframe",
  AST_POS.COL_METHOD_NAME: "method name of the column",
  AST_POS.COL_VAR: "expression representing a column",
  AST_POS.PARAM: "mandatory parameter value(s)",
  AST_POS.BINOP_RHS: "binary operator and the right-hand-side operand",
  AST_POS.COL_IDX: "column index",
  AST_POS.CELL_VALUE: "value of a cell",
  AST_POS.AGG_METHOD_NAME: "method name of the aggregation function",
  AST_POS.COMMENT: "comment written by users",
  AST_POS.CODE_LINE: "a line of code",
  AST_POS.CONDITION: "condition",
}

class PROMPT_MARKER:
  CURSOR = "<INPUT_CURSOR>"
  CODE_CONTEXT = "**code context**"
  DATA_CONTEXT = "**data context**"